/*
 * Copyright (C) 2010 Google Inc. All rights reserved.
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  Neither the name of Apple Computer, Inc. ("Apple") nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"

#if ENABLE(WEB_AUDIO)

#include "NotImplemented.h"
#include "AudioDestinationWiiU.h"
#include <wtf/Platform.h> // PLATFORM(...)
#include "AudioSourceProvider.h"
#include "AudioNode.h"	// for AudioNode::ProcessingSizeInFrames

#include <wksup/wksupAudio.h>
#include <wksup/wksupAudioPlatform.h>
#include "DefaultAudioDestinationNode.h"
#include "AudioContext.h"
#include "Frame.h"
#include "FrameLoaderClientWiiU.h"

static const unsigned kWebAudioMixingFrequency = 48000;

namespace WebCore {

class AudioDestinationWiiU::Impl : public WkSupAudioUtilsPrivate::WebAudioSampleProvider, public WkSupAudioUtilsPrivate::WebAudioRenderer
{
public:
	Impl(AudioSourceProvider &provider, unsigned mixingFreq, unsigned outputFreq)
		: m_provider(provider)
		, m_renderBus(0)
		, m_mixingFreq(mixingFreq)
		, m_outputFreq(outputFreq)
	{}

	virtual ~Impl() {
		Stop();
		if (m_renderBus)
		{
			CloseRenderer();
			delete m_renderBus;
		}
	}

	void Open(size_t nChannels, size_t nSampleRate, size_t nSampleBlockSize, AudioDestinationWiiU::OutputDevice outDevice) {
		m_renderBus = new AudioBus(nChannels, nSampleBlockSize, true);

		WebAudioRenderer::OutputDevice outDev = WebAudioRenderer::eOutDev_TV;
		switch(outDevice) {
			// Determine output from device context
			default:
			case AudioDestinationWiiU::eOutDevice_ContextDefault: {
				DefaultAudioDestinationNode *pOwner = dynamic_cast<DefaultAudioDestinationNode *>(&m_provider);
				if (pOwner) {
					AudioContext *pCtx = pOwner->context();
					if (pCtx) {
						switch(static_cast<FrameLoaderClientWiiU*>(pCtx->document()->frame()->loader()->client())->webView()->GetDisplayType()) {
							case WebView::GAMEPAD_DISPLAY_TYPE: {
								outDev = WebAudioRenderer::eOutDev_GAMEPAD;
								break;
							}
							case WebView::TV_DISPLAY_TYPE:
							case WebView::TV_LOWER_DISPLAY_TYPE:
							case WebView::TV_UPPER_DISPLAY_TYPE: {
								outDev = WebAudioRenderer::eOutDev_TV;
								break;
							}
							default:
								break;
						}
					}

				}
				break;
			}
			// Force output to TV
			case AudioDestinationWiiU::eOutDevice_TV: {
				outDev = WebAudioRenderer::eOutDev_TV;
				break;
			}
			// Force output to GAMEPAD
			case AudioDestinationWiiU::eOutDevice_GAMEPAD: {
				outDev = WebAudioRenderer::eOutDev_GAMEPAD;
				break;
			}
		}
		OpenRenderer(outDev, m_outputFreq, nChannels, nSampleRate, nSampleBlockSize, this);
	}

	void Close(void) {
		CloseRenderer();
		delete m_renderBus;
		m_renderBus = NULL;
	}

	void Start() {
		StartRenderer();
	}

	void Stop() {
		StopRenderer();
	}

private:
	virtual size_t ProvideRenderSamples(int16_t **ppSampleReceiveBuffers, size_t nChannels, size_t nSamplesRequested);

	AudioSourceProvider	&m_provider;
	AudioBus			*m_renderBus;
	unsigned			m_mixingFreq;
	unsigned			m_outputFreq;
};

size_t AudioDestinationWiiU::Impl::ProvideRenderSamples(int16_t **ppSampleReceiveBuffers, size_t nChannels, size_t nSamplesRequested) {
	// Pull samples from the pipeline.
	m_provider.provideInput(m_renderBus, nSamplesRequested);

	// Process as many channels as are possible. Any excess channels will be silenced.
	uint32_t nChMaxIn=m_renderBus->numberOfChannels();
	uint32_t nChMin = nChannels < nChMaxIn ? nChannels : nChMaxIn;
	uint32_t nCh;
	for(nCh = 0; nCh < nChMin; ++nCh) {
		const AudioChannel *pCh = m_renderBus->channel(nCh);
		int16_t *pDest = ppSampleReceiveBuffers[nCh];
		uint32_t nSmpMaxIn = pCh->length();
		uint32_t nSmpMin = nSamplesRequested < nSmpMaxIn ? nSamplesRequested : nSmpMaxIn;
		uint32_t nSmp;
		const float *pSrc = pCh->data();
		for(nSmp = 0; nSmp < nSmpMin; ++nSmp) {
			int32_t v = int32_t(*pSrc++ * 32767.0f);
			*pDest++ = v > 32767 ? 32767 : v < -32768 ? -32768 : v;
		}

		for( ; nSmp<nSamplesRequested; ++nSmp) {
			// FIXME: maybe fade out to silence
		}
	}

	for( ; nCh<nChannels; ++nCh) {
		memset(ppSampleReceiveBuffers[nCh], 0, sizeof(int16_t) * nSamplesRequested);
	}

	return(nSamplesRequested);
}

// Factory method: Create WiiU specific AudioDestination
PassOwnPtr<AudioDestination> AudioDestination::create(AudioSourceProvider& provider, double sampleRate) {
    AudioDestinationWiiU::OutputDevice outDev = AudioDestinationWiiU::eOutDevice_ContextDefault;

    return adoptPtr(new AudioDestinationWiiU(provider, sampleRate, outDev));
}

double AudioDestination::hardwareSampleRate() {
    return kWebAudioMixingFrequency;
//    return wksupAudioGetOutputSamplingRate();
}

AudioDestinationWiiU::AudioDestinationWiiU(AudioSourceProvider& provider, double sampleRate, OutputDevice outDevice)
    : m_sampleRate(sampleRate)
    , m_isPlaying(false)
 	, m_outputDevice(outDevice)
	, m_pImpl(new Impl(provider, kWebAudioMixingFrequency, wksupAudioGetOutputSamplingRate())) {
}

AudioDestinationWiiU::~AudioDestinationWiiU() {
	if (m_pImpl)
		delete m_pImpl;
}

void AudioDestinationWiiU::start() {
	if (!m_isPlaying)
		m_pImpl->Open(2, size_t(m_sampleRate), AudioNode::ProcessingSizeInFrames, m_outputDevice);
	m_pImpl->Start();
    m_isPlaying = true;
}

void AudioDestinationWiiU::stop() {
	if (m_isPlaying)
		m_pImpl->Close();
	m_pImpl->Stop();
	m_isPlaying = false;
}

} // namespace WebCore

#endif // ENABLE(WEB_AUDIO)

