/*
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions, and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY NINTENDO AND ITS CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL NINTENDO OR CONTRIBUTORS BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "GraphicsContext.h"
#include "NotImplemented.h"

#include "GraphicsContextPlatformPrivate.h"
#include <wkgfx_ClipOp.h>

#include "StatisticsWiiU.h"

#include "FontPlatformData.h"
#include "FontFallbackList.h"

#include "ShadowBlur.h"

 #include <PlatformStatistics.h>
 using namespace LibPlatform;
// --------------------------------------------------------------------------------------------------

GraphicsContextPlatformPrivate::WKgfxHooks *		GraphicsContextPlatformPrivate::s_pWKgfxHooks;

// --------------------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------------------

namespace WebCore {

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::platformInit(PlatformGraphicsContext *platformContext, int aaMode, bool useMem1, bool ignoreDepth)
{
    m_data = platformContext;
	m_isValid = m_data->init(this, aaMode, useMem1, ignoreDepth);
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::platformDestroy()
{
    delete m_data;
	m_data = NULL;
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::clipConvexPolygon(size_t numPoints, const FloatPoint* points, bool antialiased)
{
    Path p;
    p.moveTo(points[0]);
    for (size_t i = 1; i < numPoints; ++i)
        p.addLineTo(points[i]);
    clip(p); 
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::fillPath(const Path& path)
{
	if (!makeCurrent())
		return;
	FloatRect rect = path.boundingRect();

	if (m_data->isShadowed())
	{
        ShadowBlur &shadow = m_data->getShadowBlurHandler();


		if (shadow.mustUseShadowBlur(this))
		{
			IntRect bounds = clipBounds();
			if (GraphicsContext *shadowContext = shadow.beginShadowLayer(this, rect))
		{
			    shadowContext->setAlpha(getAlpha());
				shadowContext->setFillRule(fillRule());
				shadowContext->setFillColor(m_data->shadowColor(), ColorSpace());
				path.platformPath()->fillPath(shadowContext, 0);
				shadow.endShadowLayer(this);
			}
		}
		else
		{
			m_data->disableShadow();
			save();
			setFillColor(m_data->shadowColor(), ColorSpace());
			translate(m_data->shadowWidth(), m_data->shadowHeight());
			fillPath(path);
			restore();
			m_data->enableShadow();
		} 
	} 

    Gradient *g = fillGradient();
    Pattern *p = fillPattern();

    if (g)
	{
        save();
        clip(path); 
        g->fill(this, rect);
        restore();
    }
	else if (p)
	{
        Image *im = p->tileImage();
        if (im == NULL)
			return;

        FloatRect tileRect(0, 0, im->width(), im->height());
        AffineTransform affineTransform(1.0, 0.0, 0.0, 1.0, 0.0, 0.0);
        FloatPoint phase(0.0, 0.0);
        ColorSpace cs = ColorSpaceDeviceRGB;

        save();
        clip(path); 
        im->drawPattern(this, tileRect, affineTransform, phase, cs, compositeOperation(), rect);
        restore();
    }
	else
		path.platformPath()->fillPath(this, 0);
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::strokePath(const Path& path)
{
    if (strokeStyle() == NoStroke)
		return;

    if (!makeCurrent())
		return;
	
	FloatRect rect = path.boundingRect();

	// Shadowed?
	if (m_data->isShadowed())
	{
		m_data->disableShadow();
		ShadowBlur &shadow = m_data->getShadowBlurHandler();

		if (shadow.mustUseShadowBlur(this))
		{
			IntRect bounds = clipBounds();
			if (GraphicsContext *shadowContext = shadow.beginShadowLayer(this, rect))
			{
				shadowContext->setFillRule(fillRule());
				shadowContext->setAlpha(getAlpha());
				shadowContext->setStrokeColor(m_data->shadowColor(), ColorSpace());
				shadowContext->strokePath(path);
				shadow.endShadowLayer(this);
			}
		}
		else
		{
			save();
			setStrokeColor(m_data->shadowColor(), ColorSpace());
			translate(m_data->shadowWidth(), m_data->shadowHeight());
			strokePath(path);
			restore();
		}
		m_data->enableShadow();
	} 

    Gradient *g = strokeGradient();
    Pattern *p = strokePattern();

	// Any gradient?
    if (g)
	{
		// Yes. Save state...
        save();

		// Render the "strokes" into the stencil buffer
		// (in addition to possibly other clipops)
        path.platformPath()->strokeClipPath(this, NULL);

		// Make sure the stencil test is on even if we would have no ClipOps!
        m_data->enableStencilTest();
		
		// Render a full screen gradient - masked by the strokes in the stencil buffer
		g->fill(this, rect);

		// Restore things - this will also restore the stencil buffer to its old state
        restore();
    }
	// Any pattern?
	else if (p)
	{
		// Yes. It has a valid image?
        Image *im = p->tileImage();
        if (im == NULL)
			return;

		// Yes! Save context...
        save();

		// Render the stroke pattern into the stencil buffer
        // (this will - if needed - combine the data with existing ClipOp stencil information)
        path.platformPath()->strokeClipPath(this, NULL);

		// Make sure the stencil test is on even if we would have no ClipOps!
        m_data->enableStencilTest();

		// Draw the pattern
        FloatRect tileRect(0, 0, im->width(), im->height());
        AffineTransform affineTransform(1.0, 0.0, 0.0, 1.0, 0.0, 0.0);
        FloatPoint phase(0.0, 0.0);
        ColorSpace cs = ColorSpaceDeviceRGB;
        im->drawPattern(this, tileRect, affineTransform, phase, cs, compositeOperation(), rect);

		// Restore state (this includes rerendering ClipOps to restore stencil data!)
        restore();
    }
	else
	{
		// Just a simple, colored stroke. Render it directly...
        path.platformPath()->strokePath(this, 0);
	}
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::fillRect(const FloatRect& rect)
{
	if (rect.width() == 0 && rect.height() == 0)
		return;

    Gradient *g = fillGradient();
    Pattern *p = fillPattern();

    if (g)
        g->fill(this, rect);
    else if (p)
	{
        Image *im = p->tileImage();
        if (im)
		{
			FloatRect tileRect(0, 0, im->width(), im->height());
			AffineTransform affineTransform(1.0, 0.0, 0.0, 1.0, 0.0, 0.0);
			FloatPoint phase(0.0, 0.0);
			ColorSpace cs = ColorSpaceDeviceRGB; 
			im->drawPattern(this, tileRect, affineTransform, phase, cs, compositeOperation(), rect); 
		}
    }
	else
	{ 
        ColorSpace cs = ColorSpaceDeviceRGB;
        fillRect(rect, fillColor(), cs);
    }
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::clipPath(const Path& path, WindRule clipRule)
{
    clip(path);
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::drawFocusRing(const Path& path, int width, int /* offset */, const Color& color)
{
    Color prev = fillColor();
    ColorSpace cs = ColorSpaceDeviceRGB;
    float prevThickness = strokeThickness();

    if (!makeCurrent())
		return;

    setStrokeThickness(width);
    setFillColor(color, cs);

    path.platformPath()->strokePath(this, 0);

    setFillColor(prev, cs);
    setStrokeThickness(prevThickness);
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::drawFocusRing(const Vector<IntRect>& rects, int width, int /* offset */, const Color& color)
{
    Color prev = fillColor();
    ColorSpace cs = ColorSpaceDeviceRGB;
    float prevThickness = strokeThickness();

    if (!makeCurrent())
		return;

	setStrokeThickness(width);
    setFillColor(color, cs);

    for (size_t i = 0; i < rects.size(); ++i)
	{
        FloatRect fr(rects[i].x(), rects[i].y(), rects[i].width(), rects[i].height());
        strokeRect(fr, width);
    }

    setFillColor(prev, cs);
	setStrokeThickness(prevThickness);
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::drawLineForTextChecking(const FloatPoint& origin, float width, TextCheckingLineStyle style)
{
    drawLineForText(origin, width, false);
}

// --------------------------------------------------------------------------------------------------
/**
*/
FloatRect GraphicsContext::roundToDevicePixels(const FloatRect& frect, RoundingMode)
{ 
//WELL, NOW... THIS IS NOT QUITE... RIGHT! IS IT NOT? SHOULD BE RUN THROUGH THE TRANSFORMS, ROUNDED & PIPED BACK?!?
	return frect;
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::translate(float x, float y)
{
    AffineTransform a;
    a.translate(x, y);
    concatCTM(a);
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::setPlatformFillColor(const Color& col, ColorSpace colorSpace)
{
//TF: Currently the render code grabs the values directly from GraphicsContext. Hence this function is not needed.
//    One might consider a cleanup, though, to bring things into line with what the original authors seemed to intend.

    //OS::Report("fillColor: %d %d %d\n", col.red(), col.green(), col.blue());
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::setPlatformStrokeColor(const Color& col, ColorSpace colorSpace)
{
//TF: Currently the render code grabs the values directly from GraphicsContext. Hence this function is not needed.
//    One might consider a cleanup, though, to bring things into line with what the original authors seemed to intend.

    //OS::Report("strokeColor: %d %d %d\n", col.red(), col.green(), col.blue());
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::setPlatformStrokeThickness(float strokeThickness)
{
	m_data->setStrokeThickness(strokeThickness);
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::setURLForRect(const KURL& link, const IntRect& destRect)
{
    // not implemented in Cairo either
    notImplemented();
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::addInnerRoundedRectClip(const IntRect& rect, int thickness)
{
	/*
    clip(rect);
    Path p;
    FloatRect r(rect);
    // Add outer ellipse
    p.addEllipse(r);
    // Add inner ellipse
    r.inflate(-thickness);
    p.addEllipse(r);

    clip(p); 
*/
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::setPlatformShadow(FloatSize const& size, float blur, Color const& color, ColorSpace)
{
	m_data->enableShadow(size, blur, color);
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::clearPlatformShadow()
{
	m_data->disableShadow();
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::beginPlatformTransparencyLayer(float opacity)
{
	using namespace WTF; 

	if(!makeCurrent())
		return;

	m_data->beginPlatformTransparencyLayer(opacity);
	STATS_WIIU_INC_TRANSLAYER_COUNT();
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::endPlatformTransparencyLayer()
{
	using namespace WTF; 

	if (!makeCurrent())
		return;

	m_data->endPlatformTransparencyLayer();
	STATS_WIIU_DEC_TRANSLAYER_COUNT();
}

// --------------------------------------------------------------------------------------------------
/**
*/
bool GraphicsContext::supportsTransparencyLayers()
{
    return true;
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::setLineCap(LineCap lineCap)
{
    m_data->setLineCap(lineCap);
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::setLineDash(const DashArray& dashes, float dashOffset)
{
	notImplemented();
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::setLineJoin(LineJoin lineJoin)
{
    m_data->setLineJoin(lineJoin);
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::setMiterLimit(float miter)
{
    m_data->setMiterLimit(miter);
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::setAlpha(float alpha)
{
	m_data->setAlpha(alpha);
}

float GraphicsContext::getAlpha()
{
    return m_data->getAlpha();
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::setPlatformCompositeOperation(CompositeOperator op)
{
	m_data->setCompositeOperation(op);
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::clip(const Path& path)
{
    if (!makeCurrent())
		return;

	// Note: adding the clipop to the recorded list will also ensure the stencil buffer is setup correctly
    m_data->addClipOp(WKgfx::ClipOp(*path.platformPath(), m_data->getViewMatrix()));
    path.platformPath()->clipPath(this, NULL);
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::canvasClip(const Path& path)
{ 
    //cairo does this as well
    clip(path);
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::clipOut(const Path& path)
{
    if (!makeCurrent())
		return;

    WKgfx::ClipOp op(*path.platformPath(), m_data->getViewMatrix(), true);
	// Note: adding the clipop to the recorded list will also ensure the stencil buffer is setup correctly
    m_data->addClipOp(op); 
    op.apply(this); 
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::clipOut(const IntRect& r)
{
    if (!makeCurrent())
		return;

    WKgfx::ClipOp op(r.x(), r.y(), r.width(), r.height(), m_data->getViewMatrix(), true);
	// Note: adding the clipop to the recorded list will also ensure the stencil buffer is setup correctly
    m_data->addClipOp(op);
    op.apply(this); 
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::rotate(float radians)
{
    AffineTransform a;
    a.rotate(radians * (180.0f / 3.14159265358f));
    concatCTM(a);
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::scale(const FloatSize& size)
{ 
    AffineTransform a;
    a.scale(size.width(), size.height());
    concatCTM(a);
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::fillRoundedRect(const IntRect& r, const IntSize& topLeft, const IntSize& topRight, const IntSize& bottomLeft, const IntSize& bottomRight, const Color& color, ColorSpace colorSpace)
{
    if (!makeCurrent())
		return;

    Color prev = fillColor();
    setFillColor(color, colorSpace);

    if (m_data->isShadowed()) {
        ShadowBlur &shadow = m_data->getShadowBlurHandler();

        if (shadow.mustUseShadowBlur(this)) {
            if (GraphicsContext *shadowContext = shadow.beginShadowLayer(this, r)) {

                shadowContext->setFillRule(fillRule());
                shadowContext->setFillColor(m_data->shadowColor(), colorSpace);
                Path p;
                p.addRoundedRect(r, topLeft, topRight, bottomLeft, bottomRight);
                p.platformPath()->fillPath(shadowContext, 0);
                shadow.endShadowLayer(this);

            }
        } else {
            m_data->disableShadow();
            fillRoundedRect(IntRect(r.x() + m_data->shadowWidth(), r.y() + m_data->shadowHeight(), r.width(), r.height()),
                            topLeft, topRight, bottomLeft, bottomRight, m_data->shadowColor(), ColorSpace());

            m_data->enableShadow();
        }
    }
    setFillColor(color, colorSpace);
    Path p;
    p.addRoundedRect(r, topLeft, topRight, bottomLeft, bottomRight);
    p.platformPath()->fillPath(this, 0);

    setFillColor(prev, colorSpace);
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::setPlatformShouldAntialias(bool enable)
{
    //OS::Report("set platform should antialias\n");
	notImplemented();
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::setImageInterpolationQuality(InterpolationQuality quality)
{
	m_data->setImageInterpolationQuality(quality);
}

// --------------------------------------------------------------------------------------------------
/**
*/
InterpolationQuality GraphicsContext::imageInterpolationQuality() const
{
    if (m_data->imageFilterIsEnabled())
        return InterpolationDefault; 
    return InterpolationLow; 
}

// --------------------------------------------------------------------------------------------------
/**
	Make context current with graphics pipeline

	@return	True if ok, false if failed. If failed no rendering calls must be made until it returns true!
*/
bool GraphicsContext::makeCurrent()
{
	return m_data->makeCurrent();
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::invalidateBuffers()
{
	m_data->invalidateBuffers();
} 

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::fillRect(const FloatRect& rect, const Color& color, ColorSpace colorSpace)
{
	if (rect.width() == 0 && rect.height() == 0)
		return;

    if (m_data->isShadowed()) {

        ShadowBlur &shadow = m_data->getShadowBlurHandler();

        if (shadow.mustUseShadowBlur(this)) {
            if (GraphicsContext *shadowContext = shadow.beginShadowLayer(this, rect)) {
                ColorSpace cs = ColorSpaceDeviceRGB;
                shadowContext->setAlpha(getAlpha());
                shadowContext->fillRect(rect, m_data->shadowColor(), cs);
                shadow.endShadowLayer(this);
            }
        } else {
            m_data->fillRect(WebCore::FloatRect(rect.x() + m_data->shadowWidth(), rect.y() + m_data->shadowHeight(), rect.width(), rect.height()), m_state.shadowColor);
        }
    }

	m_data->fillRect(rect, color);
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::drawLineForText(const FloatPoint& origin, float width, bool printing)
{
    if (strokeStyle() == NoStroke)
		return;
	m_data->drawLineForText(origin, width, strokeColor());
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::clearRect(const FloatRect& rect)
{
	m_data->clearRect(rect); 
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::strokeRect(const FloatRect& rect, float width)
{ 
    if (strokeStyle() == NoStroke)
		return;

    WebCore::Gradient *g = strokeGradient();
    WebCore::Pattern *p = strokePattern();

    // If there is a gradient or pattern or we are drawing thick lines, we use a path instead as that will look better
    if (g || p || width >= 2.0f)
	{
        Path path;
        path.addRect(rect);
        strokePath(path); 
        return;
    }

	m_data->strokeRect(rect, width, strokeColor());
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::drawRect(const IntRect& rect)
{
	m_data->drawRect(rect, fillColor(), strokeStyle() != NoStroke, strokeColor());
}

// --------------------------------------------------------------------------------------------------
/**

	@note This is only used to draw borders, so we should not draw shadows
*/
void GraphicsContext::drawLine(const IntPoint& point1, const IntPoint& point2)
{ 
    if (strokeStyle() == NoStroke)
		return;
	m_data->drawLine(point1, point2, strokeStyle(), strokeColor());
}

// --------------------------------------------------------------------------------------------------
/**

	@note This method is only used to draw the little circles used in lists.
*/
void GraphicsContext::drawEllipse(const IntRect& rect)
{
	m_data->drawEllipse(rect, fillColor(), strokeStyle());
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::strokeArc(const IntRect& rect, int startAngle, int angleSpan)
{
    if(strokeStyle() == NoStroke)
		return;
	m_data->strokeArc(rect, startAngle, angleSpan, strokeColor(), strokeStyle());
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::drawConvexPolygon(size_t npoints, const FloatPoint* points, bool shouldAntialias)
{
	m_data->drawConvexPolygon(npoints, points, fillColor(), strokeStyle() != NoStroke, strokeColor());
}

// --------------------------------------------------------------------------------------------------
/**
*/
static const WKgfx::Matrix & _AffineTransformToMat44(WKgfx::Matrix & mat, const AffineTransform & transform)
{
	// Convert Webkit's 3x2 column-major into the HW's row-major 4x4...
	mat._00 = transform.a();
	mat._10 = transform.b();
	
	mat._01 = transform.c();
	mat._11 = transform.d();
	
	mat._02 = 0.0f;
	mat._12 = 0.0f;
	
	mat._03 = transform.e();
	mat._13 = transform.f();
	
	mat._20 = 0.0f;
	mat._21 = 0.0f;
	mat._22 = 1.0f;
	mat._23 = 0.0f;

	mat._30 = 0.0f;
	mat._31 = 0.0f;
	mat._32 = 0.0f;
	mat._33 = 1.0f;

	return mat;
}

// --------------------------------------------------------------------------------------------------
/**
*/
static const WKgfx::Matrix & _TransformationMatrixToMat44(WKgfx::Matrix & mat, const TransformationMatrix& transform)
{
    //DLM: Webkit is column major, although of course they don't mention that anywhere except on their blog

    mat._00 = transform.m11();
    mat._01 = transform.m21();
    mat._02 = transform.m31();
    mat._03 = transform.m41();

    mat._10 = transform.m12();
    mat._11 = transform.m22();
    mat._12 = transform.m32();
    mat._13 = transform.m42();

    mat._20 = transform.m13();
    mat._21 = transform.m23();
    mat._22 = transform.m33();
    mat._23 = transform.m43();

    mat._30 = transform.m14();
    mat._31 = transform.m24();
    mat._32 = transform.m34();
    mat._33 = transform.m44();


	return mat;
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::concatCTM(const AffineTransform& transform)
{
    WKgfx::Matrix mat;
	m_data->concatViewMtx(_AffineTransformToMat44(mat, transform));
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::setCTM(const AffineTransform& transform)
{ 
    WKgfx::Matrix mat;
	m_data->setViewMtx(_AffineTransformToMat44(mat, transform));
}

// --------------------------------------------------------------------------------------------------
/**
    Concatenate the transformation matrix.
    @param transform The matrix to concatenate. 
*/
void GraphicsContext::concatCTM(const TransformationMatrix &transform)
{
    WKgfx::Matrix mat;
	m_data->concatViewMtx(_TransformationMatrixToMat44(mat, transform));
}

// --------------------------------------------------------------------------------------------------
/**
    Set the current model matrix.
    @param mat The matrix to set.
*/
void GraphicsContext::setCTM(const TransformationMatrix &transform)
{
    WKgfx::Matrix mat;
	m_data->setViewMtx(_TransformationMatrixToMat44(mat, transform));
}

// --------------------------------------------------------------------------------------------------
/**
*/
AffineTransform GraphicsContext::getCTM() const
{ 
    const WKgfx::Matrix &m = m_data->getViewMatrix();
    return AffineTransform(m._00, m._10, m._01, m._11, m._03, m._13);
}


TransformationMatrix GraphicsContext::getCTMAsTransformationMatrix() const
{
    const WKgfx::Matrix &m = m_data->getViewMatrix();
    return TransformationMatrix(m._00, m._10, m._20, m._30, m._01, m._11, m._21, m._31,
                                   m._02, m._12, m._22, m._32, m._03, m._13, m._23, m._33);
}


// --------------------------------------------------------------------------------------------------
/**
 * Returns a bounding box around the current clip region. 
 * @return A bounding box around the current clip region. 
*/ 
IntRect GraphicsContext::clipBounds() const
{ 
    return m_data->clipBounds();
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::clip(const FloatRect& rect)
{ 
    if (!makeCurrent())
        return;

    const WKgfx::Matrix &m = m_data->getViewMatrix();
    //If No skew we will use scissor to clip else time to pull out the stencil buffer
    if (!m._21 && !m._20 && !m._21) {
        m_data->clipRect(rect);
    } else {
        WKgfx::ClipOp op(rect.x(), rect.y(), rect.width(), rect.height(), m, false);
        m_data->addClipOp(op);
        op.apply(this);
    }

}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::savePlatformState()
{
	m_data->pushState();
}

// --------------------------------------------------------------------------------------------------
/**
*/
void GraphicsContext::restorePlatformState()
{
    if (!makeCurrent())
		return;
	m_data->popState();
} 

} // namespace WebCore


// --------------------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------------------

// --------------------------------------------------------------------------------------------------
/**
	Startup all global systems / resources for rendering webkit content used by the graphics context
*/
void GraphicsContextPlatformPrivate::startup(WKgfx::CBpostDLCall postDLCall)
{
	s_pWKgfxHooks = new WKgfxHooks();
	ASSERT(s_pWKgfxHooks);
	WKgfx::CallbackHooks::setInstance(s_pWKgfxHooks);
	
	_startup(postDLCall);
}

// --------------------------------------------------------------------------------------------------
/**
	Shutdown all global systems / resources for rendering webkit content used by the graphics context
	
	@note Assumes the GPU to be no longer processing any resources managed by this system!
*/
void GraphicsContextPlatformPrivate::shutdown(void)
{
	endDraw();
	_shutdown();

	WKgfx::CallbackHooks::setInstance(NULL);
	delete s_pWKgfxHooks;
	s_pWKgfxHooks = NULL;
}

/*
	Implementation of the hooks needed by our WKgfx RPL to call back into Webkit
*/

void GraphicsContextPlatformPrivate::WKgfxHooks::ShadowBlur_ShadowBlur(WebCore::ShadowBlur *pThis)
{
	new(pThis) WebCore::ShadowBlur();
}

void GraphicsContextPlatformPrivate::WKgfxHooks::ShadowBlur_setShadowValues(WebCore::ShadowBlur *pThis, const WebCore::FloatSize& radius, const WebCore::FloatSize& offset, const WebCore::Color& color, WebCore::ColorSpace colorSpace, bool ignoreTransforms)
{
	pThis->setShadowValues(radius, offset, color, colorSpace, ignoreTransforms);
}

WTF::StatisticsWiiU *GraphicsContextPlatformPrivate::WKgfxHooks::StatisticsWiiU_instance(void)
{
	return WTF::StatisticsWiiU::instance();
}

void GraphicsContextPlatformPrivate::WKgfxHooks::Path_Path(WebCore::Path *pThis)
{
	new(pThis) WebCore::Path();
}

void GraphicsContextPlatformPrivate::WKgfxHooks::Path_DestructPath(WebCore::Path *pThis)
{
	pThis->WebCore::Path::~Path();
}

void GraphicsContextPlatformPrivate::WKgfxHooks::Path_addArc(WebCore::Path *pPath, const WebCore::FloatPoint& p, float radius, float startAngle, float endAngle, bool anticlockwise)
{
	pPath->addArc(p, radius, startAngle, endAngle, anticlockwise);
}

void GraphicsContextPlatformPrivate::WKgfxHooks::Path_moveTo(WebCore::Path *pPath, const WebCore::FloatPoint & endPoint)
	{
	pPath->moveTo(endPoint);
	}

void GraphicsContextPlatformPrivate::WKgfxHooks::Path_addLineTo(WebCore::Path *pPath, const WebCore::FloatPoint & endPoint)
	{
	pPath->addLineTo(endPoint);
	}

void GraphicsContextPlatformPrivate::WKgfxHooks::Path_addQuadCurveTo(WebCore::Path *pPath, const WebCore::FloatPoint & controlPoint, const WebCore::FloatPoint & endPoint)
	{
	pPath->addQuadCurveTo(controlPoint, endPoint);
	}

void GraphicsContextPlatformPrivate::WKgfxHooks::Path_addBezierCurveTo(WebCore::Path *pPath, const WebCore::FloatPoint & controlPoint1, const WebCore::FloatPoint & controlPoint2, const WebCore::FloatPoint & endPoint)
	{
	pPath->addBezierCurveTo(controlPoint1, controlPoint2, endPoint);
	}

WebCore::FloatPoint GraphicsContextPlatformPrivate::WKgfxHooks::AffineTransform_mapPoint(const WebCore::AffineTransform *pThis, const WebCore::FloatPoint& point)
{
	return pThis->mapPoint(point);
}

void GraphicsContextPlatformPrivate::WKgfxHooks::IntRect_intersect(WebCore::IntRect *pThis, const WebCore::IntRect & rect)
{
	pThis->intersect(rect);
}

void GraphicsContextPlatformPrivate::WKgfxHooks::IntRect_IntRect(WebCore::IntRect *pThis, const WebCore::FloatRect & rect)
{
	new(pThis) WebCore::IntRect(rect);
}

void GraphicsContextPlatformPrivate::WKgfxHooks::FloatRect_fitToPoints(WebCore::FloatRect *pThis, const WebCore::FloatPoint& p0, const WebCore::FloatPoint& p1, const WebCore::FloatPoint& p2, const WebCore::FloatPoint& p3)
{
	pThis->fitToPoints(p0, p1, p2, p3);
}

void GraphicsContextPlatformPrivate::WKgfxHooks::FloatRect_FloatRect(WebCore::FloatRect *pThis, const WebCore::IntRect & rect)
{
	new(pThis) WebCore::FloatRect(rect);
}

WTF::PassRefPtr<WTF::ByteArray> GraphicsContextPlatformPrivate::WKgfxHooks::ByteArray_create(size_t size)
{
	return WTF::ByteArray::create(size);
}

WTF::CString GraphicsContextPlatformPrivate::WKgfxHooks::String_ascii(const WTF::String *pThis)
{
	return pThis->ascii();
}

bool GraphicsContextPlatformPrivate::WKgfxHooks::Mutex_tryLock(Mutex *pThis)
{
	return pThis->tryLock();
}

void GraphicsContextPlatformPrivate::WKgfxHooks::Mutex_unlock(Mutex *pThis)
{
	pThis->unlock();
}

WTF::ThreadIdentifier GraphicsContextPlatformPrivate::WKgfxHooks::WTF_currentThread(void)
{
	return WTF::currentThread();
}

bool GraphicsContextPlatformPrivate::WKgfxHooks::WTF_isMainThread(void)
{
	return WTF::isMainThread();
}

const WebCore::FontData* GraphicsContextPlatformPrivate::WKgfxHooks::FontFallbackList_fontDataAt(const WebCore::FontFallbackList *pThis, const WebCore::Font* pFont, unsigned index)
{
	return pThis->fontDataAt(pFont, index);
}

WebCore::SimpleFontData* GraphicsContextPlatformPrivate::WKgfxHooks::FontCache_getCachedFontData(WebCore::FontCache *pThis, const WebCore::FontPlatformData *pData, WebCore::FontCache::ShouldRetain bRetain)
{
	return pThis->getCachedFontData(pData, bRetain);
}

void GraphicsContextPlatformPrivate::WKgfxHooks::FontPlatformData_FontPlatformData(WebCore::FontPlatformData *pThis, font_config_handle *pHandle, font_param *pParam, const WebCore::FontDescription & desc)
{
	new(pThis) WebCore::FontPlatformData(pHandle, pParam, desc);
}

void GraphicsContextPlatformPrivate::WKgfxHooks::FontPlatformData_FontPlatformData(WebCore::FontPlatformData *pThis, font_face_t *face, float size, bool bBold, bool bItalic)
{
	new(pThis) WebCore::FontPlatformData(face, size, bBold, bItalic);
}

void GraphicsContextPlatformPrivate::WKgfxHooks::FontPlatformData_FontPlatformData(WebCore::FontPlatformData *pThis, const WebCore::FontPlatformData & data)
{
	new(pThis) WebCore::FontPlatformData(data);
}

void GraphicsContextPlatformPrivate::WKgfxHooks::FontPlatformData_DestructFontPlatformData(WebCore::FontPlatformData *pThis)
{
	pThis->WebCore::FontPlatformData::~FontPlatformData();
}

void GraphicsContextPlatformPrivate::WKgfxHooks::WebKitCrash()
{
	WebKitCrash();
}

void GraphicsContextPlatformPrivate::WKgfxHooks::TimeRanges_add(WebCore::TimeRanges *pRange, float start, float end)
{
	pRange->add(start, end);
}

bool GraphicsContextPlatformPrivate::WKgfxHooks::TimeRanges_contain(const WebCore::TimeRanges *pRange, float time) const
{
	return pRange->contain(time);
}

float GraphicsContextPlatformPrivate::WKgfxHooks::TimeRanges_nearest(const WebCore::TimeRanges *pRange, float time) const
{
	return pRange->nearest(time);
}


