/*
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions, and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY NINTENDO AND ITS CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL NINTENDO OR CONTRIBUTORS BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*

	Video Playback for WiiU with Mobiclip's MVPlayer
	(high level interface with Webkit)
	
	make webcore_wiiu/fast
*/

#include <algorithm>
#include "DomainAccessControlWiiU.h"

#include "config.h"

#if 1//ENABLE(VIDEO) && PLATFORM(WIIU)

// Import wiiu's curl to define symbols like CURLcode, CURLoption ...
#include "SocketStreamHandlePrivate.h"
#include <curl/curl.h>
#include <openssl/bio.h>
#include <openssl/pem.h>
// Import WKgfx::Matrix
#include "GraphicsContextPlatformPrivate.h"
#include <wiiu/MVPlayer_wiiu.h>
#include <pthread.h>
#include <wiiu/thread_ctrl.h>


//--------------------------------------------------------------------
#include <PlatformSSL.h>
#include <PlatformDefines.h>
#include <PlatformMemory.h>
#include <PlatformNetwork.h>
#include <PlatformOS.h>
#include <PlatformError.h>
using namespace LibPlatform;

#define CURLOPT_NSSL_CONTEXT (CURLoption)210
#define CURLOPT_NSSL_PEER_VERIFY_OPT (CURLoption)211
#define NSSL_VERIFY_ALL 7

#include "wksup/wksupFS.h" // FIXME: refactor SSLRootCA read so this isn't necessary!


// --------------------------------------------------------------------


#include "MediaPlayerPrivateMobiclip.h"
#include "TimeRanges.h"

#include "Logging.h"
#include "HTMLMediaElement.h"
#include "FrameLoaderClient.h"
#include "FrameLoader.h"
#include "Frame.h"
#include "WebViewWiiU.h"
#include "FrameLoaderClientWiiU.h"

#if ENABLE_XSOCK
#include "xsockInterface.h"
#ifndef SYSTEMSLEEP
#define SYSTEMSLEEP(x) OSSleepMilliseconds(x);
#endif
#endif

#include <wtf/MainThread.h>
#include <wtf/text/StringBuilder.h>
#define LOG_MediaPlayerPrivateMobiclip 0

// --------------------------------------------------------------------
// --------------------------------------------------------------------

namespace WebCore {

// -------------------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------------------

// --------------------------------------------------------------------
/**
	Create an instance of this media engine
*/
PassOwnPtr<MediaPlayerPrivateInterface> MediaPlayerPrivateMobiclip::create(MediaPlayer* player)
{
	MediaPlayerPrivateMobiclip* o = new MediaPlayerPrivateMobiclip(player);
	// If memory allocation fails (e.g. the default heap is all taken) bail out now
	if( !o->m_v )
	{
		delete o;
		return PassOwnPtr<MediaPlayerPrivateInterface>();
	}
    return adoptPtr(o);
}

// --------------------------------------------------------------------
/**
	Return the supported MIME types of this media engine
*/
void MediaPlayerPrivateMobiclip::getSupportedTypes(HashSet<String>& types)
{
	int n = MVPlayer::GetSupportedTypesCount();
	while( n-- )
	{
		const char* stype = MVPlayer::GetSupportedType(n);
		String s(stype);
		//s = s.lower();
		if( !types.contains(s) )
			types.add(s);
	}
}

// --------------------------------------------------------------------
/**
	Check if we support a given type
*/
MediaPlayer::SupportsType MediaPlayerPrivateMobiclip::supportsType(const String& type, const String& codecs)
{
	CString ctype = type.utf8();
	const char* stype = ctype.data();
	CString ccodecs = codecs.utf8();
	const char* scodecs = ccodecs.data();

	int confidence = MVPlayer::CanPlayType( stype, scodecs );
	MediaPlayer::SupportsType supported = MediaPlayer::IsNotSupported;
	if( confidence > 0 )
	{
		supported = ( confidence > 50 ) ? MediaPlayer::IsSupported : MediaPlayer::MayBeSupported;
	}
	else
	{
		std::string typeCheck = stype;
		if(typeCheck.find("audio") != std::string::npos)
		{
			LibPlatform::Error::GetInstance()->Report(LibPlatform::Error::ERROR_CODE_WEBKIT_UNSUPPORTED_AUDIO_FORMAT, L"", 0);
		}
		else
		{
			LibPlatform::Error::GetInstance()->Report(LibPlatform::Error::ERROR_CODE_WEBKIT_UNSUPPORTED_VIDEO_FORMAT, L"", 0);
		}
	}
	return supported;
}

// --------------------------------------------------------------------
/**
	Check if we support a given type by Extension
*/
const char* MediaPlayerPrivateMobiclip::GetTypeFromExtension( const char* fileExtension )
{
	return MVPlayer::GetTypeFromExtension(fileExtension);
}


// --------------------------------------------------------------------
/**
	Register the media engine with Webkit
*/
void MediaPlayerPrivateMobiclip::registerMediaEngine(MediaEngineRegistrar registrar)
{
    registrar(create, getSupportedTypes, supportsType, 0, 0, 0);
}

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// --------------------------------------------------------------------------

class WiiUMediaPlayerObserver : public MVPlayerObserver
{
	unsigned int				m_refCount;
	MediaPlayerPrivateMobiclip* m_mediaPlayer;

public:

	WiiUMediaPlayerObserver( MediaPlayerPrivateMobiclip* mediaPlayer ):
		m_refCount(1),
		m_mediaPlayer(mediaPlayer)
	{
	}

private:

	~WiiUMediaPlayerObserver()
	{
		m_mediaPlayer = 0;
	}

public:

	unsigned int AddRef()
	{
		m_refCount ++;
		return m_refCount;
	}

	unsigned int Release()
	{
		m_refCount --;
		if( m_refCount == 0 )
		{
			delete this;
		}

		return m_refCount;
	}

public:

	void SetMediaPlayer( MediaPlayerPrivateMobiclip* mediaPlayer )
	{
		m_mediaPlayer = mediaPlayer;
	}

private:

	static void staticStateChange(void* context)
	{
		WiiUMediaPlayerObserver* This = static_cast<WiiUMediaPlayerObserver*>(context);

		if( This->m_mediaPlayer )
        {
		    This->m_mediaPlayer->OnStateChange();
        }

		This->Release();
	}

	static void staticBufferingUpdate(void* context)
	{
		WiiUMediaPlayerObserver* This = static_cast<WiiUMediaPlayerObserver*>(context);

		if( This->m_mediaPlayer )
        {
		    This->m_mediaPlayer->OnBufferingUpdate();
        }

		This->Release();
	}

	static void staticFrameUpdate(void* context)
	{
		WiiUMediaPlayerObserver* This = static_cast<WiiUMediaPlayerObserver*>(context);

		if( This->m_mediaPlayer )
        {
		    This->m_mediaPlayer->OnFrameUpdate();
        }

		This->Release();
	}

public:

	virtual void OnStateChange( MVPlayer* from, MVPlayerState newState )
	{
		AddRef();
		callOnMainThread(WiiUMediaPlayerObserver::staticStateChange, this);
	}

	virtual void OnBufferingUpdate( MVPlayer* from, float start, float end )
	{
		AddRef();
		if( from->mediaDurationSecs > 0 )
			m_mediaPlayer->setBufferingPercentage( end*100/from->mediaDurationSecs );
		callOnMainThread(WiiUMediaPlayerObserver::staticBufferingUpdate, this);
	}

	virtual void OnFrameUpdate( MVPlayer* from )
	{
		AddRef();
		callOnMainThread(WiiUMediaPlayerObserver::staticFrameUpdate, this);

	}
};

class WiiUMVHook : public MVHttpHook//, public MVThreadHook
{
	unsigned int	m_refCount;

	CString			m_userAgent;

public:

	WiiUMVHook():
		m_refCount(1),
		m_userAgent()
	{
	}

	~WiiUMVHook()
	{
	}

private:

public:

	void SetUserAgent( const CString& userAgent )
	{
		m_userAgent = userAgent;
	}

public:

	unsigned int AddRef()
	{
		m_refCount ++;
		return m_refCount;
	}

	unsigned int Release()
	{
		const unsigned ret = --m_refCount;
		if( m_refCount == 0 )
		{
			delete this;
			MV_HttpHook = 0;
		}

		return ret;
	}

public:

	int OnHttpRequest( CURL* easyRequest, const char* uri )
	{
		const char* userAgent = m_userAgent.data();
		if( userAgent )
		{
			Call_curl_easy_setopt( easyRequest, CURLOPT_USERAGENT, userAgent );
		}

		Call_curl_easy_setopt(easyRequest, CURLOPT_SSLVERSION, static_cast<long>(CURL_SSLVERSION_TLSv1));
		Call_curl_easy_setopt(easyRequest, CURLOPT_AUTOREFERER, 1L);
		Call_curl_easy_setopt(easyRequest, CURLOPT_CAINFO, LibPlatform::Network::GetCertificatePath().c_str());
		Call_curl_easy_setopt(easyRequest, CURLOPT_FAILONERROR, 1L);
		Call_curl_easy_setopt(easyRequest, CURLOPT_FOLLOWLOCATION, 1L);
		Call_curl_easy_setopt(easyRequest, CURLOPT_MAXREDIRS, 10L);
		Call_curl_easy_setopt(easyRequest, CURLOPT_PROXYAUTH, static_cast<long>(CURLAUTH_ANY));

		return 0;
	}

	void OnHttpError( CURLM* multi, CURL* easyRequest, const char* uri, curl_socket_t fd )
	{
	}
};

// --------------------------------------------------------------------------
/**
 * Returns the webview holding this document.
 * We use this to know if we are displaying on GamePad or TV.
 */
static WebView* GetWebView( HTMLMediaElement* mediaElement )
{
	Document* document = mediaElement->document();
	Frame* frame = document ? document->frame() : 0;
	FrameLoader* frameLoader = frame ? frame->loader() : 0;
	FrameLoaderClient* frameLoaderClient = frameLoader ? frameLoader->client() : 0;
	return frameLoaderClient ? static_cast<FrameLoaderClientWiiU*>(frameLoaderClient)->webView() : 0;
}

#if ENABLE_XSOCK
class MVLoadHookXsock : public MVLoadHook
{
	enum {
		PAGE_SIZE = 64 * 1024,
		MAX_PAGE_COUNT = 8
	};

	typedef std::vector<int> IdList;
	struct Page {
		int cmdId;
		int offset;
	};
	struct File {
		Page pages[MAX_PAGE_COUNT];
		int curPage;
		xsockInterface::NetworkFileHandle fd;
		xsockClient *socket;
		long long fileSize;
		long long curOffset;
		String uri;
		IdList discards;
	};
	typedef std::map<int, File> FileMap;

	int m_refCount;
	int m_nextFd;
	FileMap m_files;
	pthread_mutex_t m_filesMutex;
	char m_buffer[PAGE_SIZE];

	int getNextFd() {
		pthread_mutex_lock(&m_filesMutex);
		int ret = m_nextFd;
		if (++m_nextFd <= 0) m_nextFd = 1;
		if (m_files.count(ret)) {
			pthread_mutex_unlock(&m_filesMutex);
			return getNextFd();
		}
		pthread_mutex_unlock(&m_filesMutex);
		return ret;
	}

	int getNextPage(int curPage) {
		return (curPage + 1) % MAX_PAGE_COUNT;
	}

	long long getFileSizeLocked(File &file) {
		if (file.fileSize >= 0)
			return file.fileSize;

		xsockInterface::xsockCallbacks &callbacks = file.socket ? xsockInterface::networkCallbacks : xsockInterface::localCallbacks;
		long long result;
		if (callbacks.getFileSizeFunc) {
			if (callbacks.getFileSizeFunc(file.socket, file.uri.utf8().data(), &file.fileSize))
				return file.fileSize;
		}
		return -1;
	}

	int requestNextPage(File &file, long long offset) {
		if (!file.socket) return 0;

		long long remainingSize = getFileSizeLocked(file);
		remainingSize -= offset;
		if (remainingSize < 0) return -1;

		if (offset != file.curOffset) {
			discardAllPages(file);
			if (xsockInterface::networkCallbacks.seekFileFunc) {
				long long resultSeek;
				if ((resultSeek = xsockInterface::networkCallbacks.seekFileFunc(file.socket, file.fd, offset, 0)) != offset) {
					return -1;
				}
			}
			else return -1;
		}
		int firstAvailablePage = file.curPage;
		while (file.pages[firstAvailablePage].cmdId) {
			firstAvailablePage = getNextPage(firstAvailablePage);
			if (firstAvailablePage == file.curPage) return 0;
		}
		int numAvailablePages = (file.curPage - firstAvailablePage + MAX_PAGE_COUNT) % MAX_PAGE_COUNT;
		if (numAvailablePages == 0) numAvailablePages = MAX_PAGE_COUNT;
		int sizeToRead = numAvailablePages * PAGE_SIZE;
		if (remainingSize < sizeToRead) sizeToRead = remainingSize;
		int numPagesUsed = sizeToRead / PAGE_SIZE + (sizeToRead % PAGE_SIZE ? 1 : 0);
		if (numPagesUsed == 0) return 0;
		if (xsockInterface::networkCallbacks.readFromFileNetwork_asyncSend) {
			int cmd = xsockInterface::networkCallbacks.readFromFileNetwork_asyncSend(file.socket, file.fd, sizeToRead, PAGE_SIZE);
			if (cmd < 0) return -1;
			for (int i = 0; i < numPagesUsed; ++i) {
				file.pages[firstAvailablePage].cmdId = cmd + i;
				file.pages[firstAvailablePage].offset = offset + i * PAGE_SIZE;
				firstAvailablePage = getNextPage(firstAvailablePage);
			}
			file.curOffset = offset + sizeToRead;
			return numPagesUsed;
		}
		else return -1;
	}

	int discardPage(File &file, int page) {
		if (file.pages[page].cmdId == 0) return 0;
		file.discards.push_back(file.pages[page].cmdId);
		file.pages[page].cmdId = 0;
		file.pages[page].offset = 0;
		return 1;
	}

	void discardAllPages(File &file) {
		while (discardPage(file, file.curPage)) file.curPage = getNextPage(file.curPage);
	}

	int clearDiscardedPages(File &file) {
		if (!file.socket) return 0;
		size_t numDiscarded = 0;
		for (numDiscarded = 0; numDiscarded < file.discards.size(); ++numDiscarded) {
			int result = xsockInterface::networkCallbacks.readFromFileNetwork_asyncRecv(file.socket, file.discards[numDiscarded], NULL);
			if (result < 0) {
				if (result == xsockInterface::operationWouldBlock) break;
				file.discards.clear();
				return 0;
			}
		}
		if (numDiscarded) file.discards.erase(file.discards.begin(), file.discards.begin() + numDiscarded);

		return (int)file.discards.size();
	}

public:

	int AddRef() {
		pthread_mutex_lock(&m_filesMutex);
		int ret = ++m_refCount;
		pthread_mutex_unlock(&m_filesMutex);
		return ret;
	}

	int Release() {
		pthread_mutex_lock(&m_filesMutex);
		int ret = --m_refCount;
		pthread_mutex_unlock(&m_filesMutex);
		return ret;
	}

public:
	
	MVLoadHookXsock() : m_refCount(0), m_nextFd(1) {
		pthread_mutex_init(&m_filesMutex, NULL);
		AddRef();
	}

	~MVLoadHookXsock() {
		while (!m_files.empty()) {
			int fd = m_files.begin()->first;
			Close(fd);
		}
		pthread_mutex_destroy(&m_filesMutex);
	}

	virtual long long GetFileSize(int fd) {
		pthread_mutex_lock(&m_filesMutex);
		if (!m_files.count(fd)) {
			pthread_mutex_unlock(&m_filesMutex);
			return -1;
		}

		File &file = m_files[fd];
		long long ret = getFileSizeLocked(file);
		pthread_mutex_unlock(&m_filesMutex);
		return ret;
	}

	virtual int OpenURI(const char *uri) {
		xsockClient *socket = NULL;
		if (xsockInterface::localCallbacks.isNetworkHostFunc)
			xsockInterface::localCallbacks.isNetworkHostFunc("", &socket);
		xsockInterface::xsockCallbacks &callbacks = socket ? xsockInterface::networkCallbacks : xsockInterface::localCallbacks;
		int fd = getNextFd();
		pthread_mutex_lock(&m_filesMutex);

		File &file = m_files[fd];
		memset(&file.pages, 0, sizeof(Page) * MAX_PAGE_COUNT);
		file.curPage = 0;
		file.fd = -1;
		file.socket = socket;
		file.fileSize = -1;
		file.curOffset = 0;
		file.uri = uri;

		if (callbacks.openFileFunc)
			file.fd = callbacks.openFileFunc(socket, uri, "r");
		if (file.fd < 0) {
			m_files.erase(fd);
			fd = -1;
		}
		else {
			requestNextPage(file, 0);
		}

		pthread_mutex_unlock(&m_filesMutex);
		return fd;
	}

	virtual int Read(int fd, long long offset, void *buffer, int count) {
		pthread_mutex_lock(&m_filesMutex);
		if (!m_files.count(fd)) {
			pthread_mutex_unlock(&m_filesMutex);
			return 0;
		}
		File &file = m_files[fd];
		int sizeRead = 0;
		if (file.socket) {
			clearDiscardedPages(file);
			if (file.pages[file.curPage].cmdId && file.pages[file.curPage].offset == offset) {
				int read = xsockInterface::operationWouldBlock;
				while (read == xsockInterface::operationWouldBlock) {
					read = xsockInterface::networkCallbacks.readFromFileNetwork_asyncRecv(file.socket, file.pages[file.curPage].cmdId, m_buffer);
					if (read == xsockInterface::operationWouldBlock) {
						SYSTEMSLEEP(1);
					}
				}
				file.pages[file.curPage].cmdId = 0;
				file.pages[file.curPage].offset = 0;
				file.curPage = getNextPage(file.curPage);
				if (read > 0) {
					if (count < read) read = count;
					memcpy(buffer, (void*)m_buffer, read);
					sizeRead = read;
					requestNextPage(file, file.curOffset);
				}
			}
			else {
				discardAllPages(file);
				if (requestNextPage(file, offset) > 0) {
					pthread_mutex_unlock(&m_filesMutex);
					return Read(fd, offset, buffer, count);
				}
			}
		}
		else {
			if (file.curOffset != offset) {
				if (xsockInterface::localCallbacks.seekFileFunc) {
					file.curOffset = xsockInterface::localCallbacks.seekFileFunc(file.socket, file.fd, offset, 0);
				}
				if (file.curOffset != offset) {
					pthread_mutex_unlock(&m_filesMutex);
					return 0;
				}
			}

			if (xsockInterface::localCallbacks.readFromFileFunc) {
				while (count > 0) {
					int result = xsockInterface::localCallbacks.readFromFileFunc(file.socket, file.fd, ((char*)buffer) + sizeRead, count);
					if (result < 0) break;
					count -= result;
					sizeRead += result;
				}
				file.curOffset += sizeRead;
			}
		}
		pthread_mutex_unlock(&m_filesMutex);
		return sizeRead;
	}

	virtual void Close(int fd) {
		pthread_mutex_lock(&m_filesMutex);
		if (!m_files.count(fd)) {
			pthread_mutex_unlock(&m_filesMutex);
			return;
		}

		File &file = m_files[fd];
		xsockInterface::xsockCallbacks &callbacks = file.socket ? xsockInterface::networkCallbacks : xsockInterface::localCallbacks;
		if (callbacks.closeFileFunc)
			callbacks.closeFileFunc(file.socket, &file.fd);
		discardAllPages(file);
		while (clearDiscardedPages(file));
		callbacks.closeNetworkHostFunc(file.socket);
		m_files.erase(fd);

		pthread_mutex_unlock(&m_filesMutex);
	}

};
#endif

// --------------------------------------------------------------------------
/**
*/
MediaPlayerPrivateMobiclip::MediaPlayerPrivateMobiclip(MediaPlayer* player)
	: m_iSeekPerformed(false)
	, m_fgCurrentTimeSecs(0.0f)
	, m_bufferingPercentage(0)
	, m_isRemote(false)
	, m_isLocal(false)
	, m_isInBackground(false)
	, m_player(player)
	, m_previousStatePreparing(false)
	, m_preLoadMode(MediaPlayer::Auto)
	, m_networkState(MediaPlayer::Empty)
	, m_readyState(MediaPlayer::HaveNothing)
	, m_size(IntSize())
{
	m_v = MVPlayer::Create();
	m_vObserver = new WiiUMediaPlayerObserver(this);
	ProcessSwitchStart();

	if( MV_HttpHook == 0 )
	{
		MV_HttpHook = new WiiUMVHook;
	}
	else
	{
		static_cast<WiiUMVHook*>(MV_HttpHook)->AddRef();
	}
	#if ENABLE_XSOCK
	if (MV_LoadHook == 0)
		MV_LoadHook = new MVLoadHookXsock;
	else
		static_cast<MVLoadHookXsock*>(MV_LoadHook)->AddRef();
	#endif
}

// --------------------------------------------------------------------------
/**
*/
MediaPlayerPrivateMobiclip::~MediaPlayerPrivateMobiclip()
{
	ProcessSwitchEnd();
	m_v->observer = 0;
	static_cast<WiiUMediaPlayerObserver*>(m_vObserver)->SetMediaPlayer( 0 );
	static_cast<WiiUMediaPlayerObserver*>(m_vObserver)->Release();

	static_cast<WiiUMVHook*>(MV_HttpHook)->Release();

	MVPlayer::Destroy( m_v );

	#if ENABLE_XSOCK
	if (static_cast<MVLoadHookXsock*>(MV_LoadHook)->Release() == 0) {
		delete MV_LoadHook;
		MV_LoadHook = 0;
	}
	#endif
}

// --------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateMobiclip::setPreload(MediaPlayer::Preload preLoad)
{
	m_preLoadMode = preLoad;
}

// --------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateMobiclip::load(const String& url)
{
	// If the url of the video being loaded is from a non-whitelist site,
	// then return and don't allow the load
	// This should probably be done  somewhere else...
	int error = WebCore::DomainAccessControl::DOMAIN_ACCESS_OK;
	if(!WebCore::DomainAccessControl::IsDomainAccessible(url, &error)) {
		switch(error) {
			case WebCore::DomainAccessControl::DOMAIN_ACCESS_NOT_ACQUIRED_AUTH_TOKEN:
				OS::Report("Media player cannot load! Needs authentication token.\n"); 
				break;

			case WebCore::DomainAccessControl::DOMAIN_ACCESS_NOT_WHITELISTED: 
				OS::Report("Media player cannot load! Domain access not whitelisted.\n"); 
				LibPlatform::Error::GetInstance()->Report(LibPlatform::Error::ERROR_CODE_WEBKIT_ADDRESS_BLOCKED_BY_WHITELIST, L"", 0);
				break;
		}
		return;
	}

	m_mediaURL = url;

	CString curl = url.utf8();
	const char* surl = curl.data();

	m_networkState	= MediaPlayer::Empty;
	m_readyState	= MediaPlayer::HaveNothing;

	m_isRemote		= false;
	m_isLocal		= false;
	m_loadType		= MediaPlayer::Unknown;

	if( strncmp(surl, "http", 4) == 0 )
	{
		HTMLMediaElement* mediaElement = static_cast<HTMLMediaElement*>( m_player->mediaPlayerClient() );
		Document* document = mediaElement->document();
		Frame* frame = document ? document->frame() : 0;
		FrameLoader* frameLoader = frame ? frame->loader() : 0;

		if( frameLoader )
		{
			const KURL& src = mediaElement->currentSrc();

			CString userAgent = frameLoader->userAgent(src).utf8();

			static_cast<WiiUMVHook*>(MV_HttpHook)->SetUserAgent( userAgent );
		}
		m_isRemote = true;
	}
	else
	{
		m_isLocal = (strncmp(surl, "file", 4) == 0);
		if( m_isLocal )
		{
			m_loadType = MediaPlayer::StoredStream;
		}
	}

	m_bufferingPercentage = 0;

	m_v->Abort();	// Make sure to close any previously playing media on this instance
	m_v->SetSource( surl );
	m_v->observer = m_vObserver;

	m_v->viewTV.volumePercent = 0;
	m_v->viewGamePad.volumePercent = 0;

	if( m_preLoadMode != MediaPlayer::None )
	{
#if LOG_MediaPlayerPrivateMobiclip==1
		OS::Report( "[MediaPlayerPrivateMobiclip] AutoStart\n" );
#endif
		m_v->Prepare();
	}
}


// --------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateMobiclip::cancelLoad()
{
	m_v->Abort();
	newNetworkState(MediaPlayer::Empty);
	newReadyState(MediaPlayer::HaveNothing);
}

// --------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateMobiclip::prepareToPlay()
{
#if LOG_MediaPlayerPrivateMobiclip==1
	OS::Report( "[MediaPlayerPrivateMobiclip] prepareToPlay\n" );
#endif
	m_v->Prepare();
}

#if USE(ACCELERATED_COMPOSITING)
// --------------------------------------------------------------------------
/**
*/
PlatformLayer* MediaPlayerPrivateMobiclip::platformLayer() const
{
	return m_videoLayer.get();
}
#endif

// --------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateMobiclip::play()
{
#if LOG_MediaPlayerPrivateMobiclip==1
	OS::Report( "[MediaPlayerPrivateMobiclip] play\n" );
#endif
	m_v->Play();
}

// --------------------------------------------------------------------------
/**
	Pause playback
*/
void MediaPlayerPrivateMobiclip::pause()
{
	bool isPlaying = (m_v->state == MVPLAYER_STATE_PLAYING);
#if LOG_MediaPlayerPrivateMobiclip==1
	OS::Report( "[MediaPlayerPrivateMobiclip] pause (forwarded:%d)\n", isPlaying );
#endif
	if( isPlaying )
		m_v->Pause();
}

// --------------------------------------------------------------------------
/**
*/
IntSize MediaPlayerPrivateMobiclip::naturalSize() const
{
    if (!hasVideo() || m_readyState < MediaPlayer::HaveMetadata)
        return IntSize();
    return IntSize(m_v->videoWidth, m_v->videoHeight);
}

// --------------------------------------------------------------------------
/**
*/
bool MediaPlayerPrivateMobiclip::hasVideo() const
{
	return m_v && (m_v->videoWidth > 0);
}

// --------------------------------------------------------------------------
/**
*/
bool MediaPlayerPrivateMobiclip::hasAudio() const
{
	return m_v && (m_v->audioFrequency > 0);
}

// --------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateMobiclip::setVisible(bool bVisible)
{
	WebView* webView = GetWebView( static_cast<HTMLMediaElement*>(m_player->mediaPlayerClient()) );

	MVView* playerView = &m_v->viewTV; // By default render to TV
	if( webView )
	{
		// If we can query information on display device, make a better choice
		playerView = webView->GetDisplayType() == WebView::TV_DISPLAY_TYPE ? &m_v->viewTV : &m_v->viewGamePad;
	}

	playerView->visibility = (bVisible ? MVPLAYER_VISIBILITY_VIDEO_ONLY : MVPLAYER_VISIBILITY_HIDDEN);
#if USE(ACCELERATED_COMPOSITING)    
    if (bVisible)
        setUpVideoRendering();
#endif
}

//---------------------------------------------------------------------------
/**
*/
float MediaPlayerPrivateMobiclip::duration() const
{
	return (float)m_v->mediaDurationSecs;
}



//---------------------------------------------------------------------------
/**
	Get current time of playback in seconds
	
	@note	This will return 0.0 if no data is known
*/
float MediaPlayerPrivateMobiclip::currentTime() const
{
	float currentTime = (float)m_v->GetCurrentTimeSecs();
#if LOG_MediaPlayerPrivateMobiclip==1
	OS::Report( "[MediaPlayerPrivateMobiclip] currentTime %f\n", currentTime );
#endif
	return currentTime;
}



//---------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateMobiclip::seek(float time)
{
#if LOG_MediaPlayerPrivateMobiclip==1
	OS::Report( "[MediaPlayerPrivateMobiclip] seek %f\n", time );
#endif
	m_v->SeekTo( time );

	if( m_v->state != MVPLAYER_STATE_SEEKING )
	{
		// We didn't seek because data were already available
		newReadyState(MediaPlayer::HaveCurrentData);
	}
}

//---------------------------------------------------------------------------
/**
*/
bool MediaPlayerPrivateMobiclip::seeking() const
{
#if LOG_MediaPlayerPrivateMobiclip==1
	OS::Report( "[MediaPlayerPrivateMobiclip] seeking\n" );
#endif
    return m_v->state == MVPLAYER_STATE_SEEKING;
}

//---------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateMobiclip::setRate(float rate)
{
	const int res = m_v->SetPlaybackRate(rate);
	OS::Report("MVPlayer::SetPlaybackRate(%f) returned %d\n", rate, res);
}

//---------------------------------------------------------------------------
/**
*/
bool MediaPlayerPrivateMobiclip::paused() const
{
	bool isPaused = (m_v->state != MVPLAYER_STATE_PLAYING);
	//bool isPaused = (m_v->state == MVPLAYER_STATE_PAUSED);
#if LOG_MediaPlayerPrivateMobiclip==1
	OS::Report( "[MediaPlayerPrivateMobiclip] paused %d\n", isPaused );
#endif
    return isPaused;
}

//---------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateMobiclip::setVolume(float pcent)
{
	WebView* webView = GetWebView( static_cast<HTMLMediaElement*>(m_player->mediaPlayerClient()) );

	MVView* playerView = &m_v->viewTV; // By default render to TV
	MVView* otherView = &m_v->viewGamePad;
	if( webView )
	{
		// If we can query information on display device, make a better choice
		if( webView->GetDisplayType() != WebView::TV_DISPLAY_TYPE )
		{
			std::swap(playerView, otherView);
		}
	}

	playerView->volumePercent = pcent*100;
	// Whatever display device is not active should be muted
	otherView->volumePercent = 0;
}

//---------------------------------------------------------------------------
/**
*/
MediaPlayer::NetworkState MediaPlayerPrivateMobiclip::networkState() const
{
#if LOG_MediaPlayerPrivateMobiclip==1
	OS::Report( "[MediaPlayerPrivateMobiclip] networkState %d\n", m_networkState );
#endif
    return(m_networkState);
}

//---------------------------------------------------------------------------
/**
*/
MediaPlayer::ReadyState MediaPlayerPrivateMobiclip::readyState() const
{
#if LOG_MediaPlayerPrivateMobiclip==1
	OS::Report( "[MediaPlayerPrivateMobiclip] readyState %d\n", m_readyState );
#endif
    return(m_readyState);
}

//-----------------------------------------------------------------------------
/**
*/
float MediaPlayerPrivateMobiclip::maxTimeSeekable() const
{
	return duration();
}

//---------------------------------------------------------------------------
/**
*/
PassRefPtr<TimeRanges> MediaPlayerPrivateMobiclip::buffered() const
{
    RefPtr<TimeRanges> timeRanges = TimeRanges::create();
	float start, end;
	m_v->GetBufferedRange(start,end);
#if LOG_MediaPlayerPrivateMobiclip==1
	OS::Report( "[MediaPlayerPrivateMobiclip] buffered Range %f %f\n", start, end );
#endif
    timeRanges->add( start, end );
    return timeRanges.release();
}

//---------------------------------------------------------------------------
/**
*/
unsigned MediaPlayerPrivateMobiclip::bytesLoaded() const
{
	return m_v->GetBytesLoaded();
}

//---------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateMobiclip::setSize(const IntSize& size)
{
    m_size = size;
#if USE(ACCELERATED_COMPOSITING)
		if( m_videoLayer.get() && supportsAcceleratedRendering() )
			m_videoLayer->setNeedsDisplay( IntRect( 0, 0, size.width(), size.height()));
#endif
}

//---------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateMobiclip::paint(GraphicsContext* context, const IntRect& rect)
{
	// Please see comment at method: MediaPlayerPrivateMobiclip::naturalSize()
	if (context->paintingDisabled() || !hasVideo())
        return;

	if( !rect.isEmpty() )
	{
		GraphicsContextPlatformPrivate* g = context->platformContext();

		WebView* webView = GetWebView( static_cast<HTMLMediaElement*>(m_player->mediaPlayerClient()) );

		MVView* playerView = &m_v->viewTV; // By default render to TV
		if( webView )
		{
			// If we can query information on display device, make a better choice
			playerView = webView->GetDisplayType() == WebView::TV_DISPLAY_TYPE ? &m_v->viewTV : &m_v->viewGamePad;
		}

		playerView->AdvanceFrame();
		MVPlayerDecodedImage image;
		if( playerView->GetDecodedImage( &image ) )
		{
			FloatRect src(image.left, image.top, image.right, image.bottom);
			g->drawDataYUV(FloatRect(rect), src, const_cast<void*>(image.y), const_cast<void*>(image.uv), image.width, image.height);
			*image.submittedTimeStamp = getSubmittedTimeStamp();
		}
		else
		{
			// No data. A simple black rectangle must do...
			Color black(0,0,0,255);
			context->fillRect(FloatRect(rect), black, ColorSpaceDeviceRGB);
		}
	}
}

//-----------------------------------------------------------------------------
/**
*/
bool MediaPlayerPrivateMobiclip::hasSingleSecurityOrigin() const
{
    // We tell to disallow resources that come from different origins so we are sure all media is single origin
    return true;
}

//-----------------------------------------------------------------------------
/**
*/
bool MediaPlayerPrivateMobiclip::newReadyState(MediaPlayer::ReadyState newState)
{
	bool bRet = (m_readyState != newState);

	if( bRet )
	{
		m_readyState = newState;
		m_player->readyStateChanged();
	}

	return(bRet);
}

//-----------------------------------------------------------------------------
/**
*/
bool MediaPlayerPrivateMobiclip::newNetworkState(MediaPlayer::NetworkState newState)
{
	bool bRet = (m_networkState != newState);

	if( bRet )
	{
		m_networkState = newState;
		m_player->networkStateChanged();
	}

	return(bRet);
}

//-----------------------------------------------------------------------------
/**
*/
MediaPlayer::MovieLoadType MediaPlayerPrivateMobiclip::movieLoadType() const
{
	//Unknown, Download, StoredStream, LiveStream 
#if LOG_MediaPlayerPrivateMobiclip==1
	OS::Report( "[MediaPlayerPrivateMobiclip] movieLoadType %d\n", m_loadType );
#endif

	return m_loadType;
}


#if USE(ACCELERATED_COMPOSITING)

//-----------------------------------------------------------------------------
/**
*/ 
// whether accelerated rendering is supported by the media engine for the current media.
bool MediaPlayerPrivateMobiclip::supportsAcceleratedRendering() const
{
    return true;
}

//-----------------------------------------------------------------------------
/**
*/ 
// called when the rendering system flips the into or out of accelerated rendering mode.
void MediaPlayerPrivateMobiclip::acceleratedRenderingStateChanged()
{ 
    setUpVideoRendering(); 
} 

bool MediaPlayerPrivateMobiclip::isReadyForVideoSetup() const
{
    return m_readyState >= MediaPlayer::HaveMetadata && m_player->visible();
}

MediaPlayerPrivateMobiclip::MediaRenderingMode MediaPlayerPrivateMobiclip::currentRenderingMode() const
{
    if (platformLayer())
       return MediaRenderingToLayer;

    if (m_player->mediaPlayerClient())
        return MediaRenderingToContext;

    return MediaRenderingNone;
}

MediaPlayerPrivateMobiclip::MediaRenderingMode MediaPlayerPrivateMobiclip::preferredRenderingMode() const
{
    if (!m_player->visible() || !m_player->frameView())
        return MediaRenderingNone;

    if (supportsAcceleratedRendering() && m_player->mediaPlayerClient()->mediaPlayerRenderingCanBeAccelerated(m_player))
         return MediaRenderingToLayer;

    return MediaRenderingToContext;
}

void MediaPlayerPrivateMobiclip::setUpVideoRendering()
{
    if (!isReadyForVideoSetup())
        return;

    MediaRenderingMode currentMode = currentRenderingMode();
    MediaRenderingMode preferredMode = preferredRenderingMode();
    if (currentMode == preferredMode && currentMode != MediaRenderingNone)
       return;

    if (currentMode != MediaRenderingNone) {  
        tearDownVideoRendering();
    }
    if (preferredMode == MediaRenderingToLayer) {
        createVideoLayer();   
    }

    // If using a movie layer, inform the client so the compositing tree is updated.
    if (currentMode == MediaRenderingToLayer || preferredMode == MediaRenderingToLayer) {
          m_player->mediaPlayerClient()->mediaPlayerRenderingModeChanged(m_player);
   }

}

void MediaPlayerPrivateMobiclip::createVideoLayer()
{
    if (m_videoLayer) {
        return;
    }
     m_videoLayer = VideoLayerWiiU::create(0); 
     m_videoLayer->setPlayer(m_player); 
}

void MediaPlayerPrivateMobiclip::tearDownVideoRendering()
{
    if (platformLayer()) {
        m_videoLayer.release(); 
    }
}
#endif
//---------------------------------------------
//---------------------------------------------
//---------------------------------------------

void MediaPlayerPrivateMobiclip::OnFrameUpdate()
{
	m_player->repaint();

#if USE(ACCELERATED_COMPOSITING)
		if( m_videoLayer.get() && supportsAcceleratedRendering() )
			m_videoLayer->setNeedsDisplay( IntRect( 0, 0, m_videoLayer->bounds().width(), m_videoLayer->bounds().height()) );
#endif

	float startSecs, endSecs;
	m_v->GetBufferedRange( startSecs, endSecs );
	m_player->timeChanged();
	if( (m_v->mediaDurationSecs > 0) && (endSecs == m_v->mediaDurationSecs) )
	{
		m_bufferingPercentage = 100;
		OnBufferingUpdate();
	}
}

void MediaPlayerPrivateMobiclip::OnBufferingUpdate()
{
	float startSecs=0, endSecs=0;
	m_v->GetBufferedRange( startSecs, endSecs );

	if( (m_bufferingPercentage == 100 || endSecs-startSecs > 3) && (m_v->state == MVPLAYER_STATE_PAUSED || m_v->state == MVPLAYER_STATE_PLAYING) )
	{
/*
		Removed since this has an undesired side effect: telling that everything is loaded prevents
		an instance of our class to be destroyed!

		This is caused by (a bug in?) WebCore::HTMLMediaElement::userCancelledLoad() where they leave
		the function when the file has been completely loaded:
			if (m_networkState == NETWORK_EMPTY || m_completelyLoaded)

		Since the function is left early the player is not destroyed a couple of lines later:
			m_player.clear();

		newNetworkState(MediaPlayer::Loaded);
*/
		newReadyState(MediaPlayer::HaveEnoughData); // Enough to reach the end without stalling
	}
}

void MediaPlayerPrivateMobiclip::OnStateChange()
{
	MVPlayerState state = m_v->state;
#if LOG_MediaPlayerPrivateMobiclip==1
	OS::Report( "[MediaPlayerPrivateMobiclip] OnStateChange %d\n", state );
#endif

	if( m_isInBackground )
	{
		// We are recovering from background mode
		// The video is loading and it will switch to paused state once it's done
		// Then we have to seek to the saved playback position and wait again for paused state (i.e. end of seek)
		// see ProcessSwitchResumeState
		if( state == MVPLAYER_STATE_PAUSED )
		{
			if( m_iSeekPerformed )
			{
				// We are at the right playback position
				// Mobiclip's media player is ready to resume
				// Tell webkit's HTMLMediaElement to quit freeze state
				HTMLMediaElement* mediaElement = static_cast<HTMLMediaElement*>( m_player->mediaPlayerClient() );
				Document* document = mediaElement->document();
				document->resumeActiveDOMObjects();
				m_isInBackground = false;
			}
			else
			{
				if( m_fgCurrentTimeSecs > 0.0f )
				{
					seek(m_fgCurrentTimeSecs);
				}
				m_iSeekPerformed = true;
			}
		}
	}
	else
	{
		switch( state )
		{
		case MVPLAYER_STATE_SEEKING:
		case MVPLAYER_STATE_BUFFERING:
			newReadyState(MediaPlayer::HaveCurrentData);	// Data at playback position is available
			break;

		case MVPLAYER_STATE_PLAYING:
		case MVPLAYER_STATE_PAUSED:
		case MVPLAYER_STATE_COMPLETE:
			if (state == MVPLAYER_STATE_COMPLETE)
			{
				// When the audio stream is longer than the video stream, the HTML "ended" event isn't fired.
				// This is because the "timeChanged" function is only called in OnFrameUpdate.
				// timeChanged calls HTMLMediaElement::mediaPlayerTimeChanged, which is where the ended event is fired from.
				OS::Report("Media ended, forcing one last \"timeChanged\" update (%f, %f)\n", currentTime(), duration());
				m_player->timeChanged();
			}
			if (m_previousStatePreparing == true) 
			{
				//We now have the meta data 
				newReadyState(MediaPlayer::HaveMetadata);	
				m_previousStatePreparing = false; 
			}
			if( m_isRemote )
			{
				m_loadType = (duration() > 0.0) ? MediaPlayer::Download : MediaPlayer::LiveStream;
			}

			if( m_loadType == MediaPlayer::LiveStream || m_loadType == MediaPlayer::StoredStream )
			{
/*
				Removed since this has an undesired side effect: telling that everything is loaded prevents
				an instance of our class to be destroyed!

				This is caused by (a bug in?) WebCore::HTMLMediaElement::userCancelledLoad() where they leave
				the function when the file has been completely loaded:
					if (m_networkState == NETWORK_EMPTY || m_completelyLoaded)

				Since the function is left early the player is not destroyed a couple of lines later:
					m_player.clear();

				newNetworkState(MediaPlayer::Loaded);
*/
				newReadyState(MediaPlayer::HaveEnoughData); // Enough to reach the end without stalling
			}
			else
			{
				newReadyState( MediaPlayer::HaveFutureData);	// Have data to play for a while
			}
			break;

		case MVPLAYER_STATE_PREPARING:
			newNetworkState(MediaPlayer::Loading);
			m_previousStatePreparing = true; 
			break;
		}
	}

	if( state == MVPLAYER_STATE_ERROR )
	{
		newReadyState(MediaPlayer::HaveNothing);
		switch(m_v->error)
		{
			case MVPLAYER_ERROR_ABORTED:
			case MVPLAYER_ERROR_NETWORK:
				newNetworkState(MediaPlayer::NetworkError);
				break;
			case MVPLAYER_ERROR_DECODE:
				newNetworkState(MediaPlayer::DecodeError);
				break;
			case MVPLAYER_ERROR_SRC_NOT_SUPPORTED:
				newNetworkState(MediaPlayer::FormatError);
				break;
		}
	}

}

int MediaPlayerPrivateMobiclip::ProcessSwitchStart()
{
	pthread_mutex_init( &m_ProcessSwitchMutex, NULL );
	pthread_cond_init( &m_ProcessSwitchCond, NULL );

	pthread_attr_t attr;
	pthread_attr_init( &attr );
	pthread_attr_setthreadname( &attr, "Mobiclip ProcessSwitchLoop" );
	sched_param_t sparam;
	sparam.sched_priority = PTHREAD_WIIU_PRIO_DEFAULT;
	sparam.sched_affinity = PTHREAD_WIIU_AFFINITY_CORE0;
	pthread_attr_setschedparam( &attr, &sparam );
	pthread_create( &m_ProcessSwitchThread, &attr, ProcessSwitchLoopStatic, this );
}

void MediaPlayerPrivateMobiclip::ProcessSwitchEnd()
{
	pthread_cond_signal( &m_ProcessSwitchCond );
	void* ignore;
	pthread_join( m_ProcessSwitchThread, &ignore );

	pthread_mutex_destroy( &m_ProcessSwitchMutex );
	pthread_cond_destroy( &m_ProcessSwitchCond );
}

void* MediaPlayerPrivateMobiclip::ProcessSwitchLoopStatic( void* arg )
{
	((MediaPlayerPrivateMobiclip*)arg)->ProcessSwitchLoop();
	return 0;
}

void MediaPlayerPrivateMobiclip::ProcessSwitchPauseStatic(void* context)
{
	MediaPlayerPrivateMobiclip* const This = static_cast<MediaPlayerPrivateMobiclip*>( context );
    //~ pthread_mutex_lock( &This->m_ProcessSwitchMutex );

	MediaPlayerClient* const mpc = This->m_player->mediaPlayerClient();
	WebCore::HTMLMediaElement* const hme = static_cast<WebCore::HTMLMediaElement*>( mpc );
	hme->pause();

	//~ pthread_cond_signal( &This->m_ProcessSwitchCond );
    //~ pthread_mutex_unlock( &This->m_ProcessSwitchMutex );
}

void MediaPlayerPrivateMobiclip::ProcessSwitchPlayStatic(void* context)
{
	MediaPlayerPrivateMobiclip* const This = static_cast<MediaPlayerPrivateMobiclip*>( context );
    //~ pthread_mutex_lock( &This->m_ProcessSwitchMutex );

	MediaPlayerClient* const mpc = This->m_player->mediaPlayerClient();
	WebCore::HTMLMediaElement* const hme = static_cast<WebCore::HTMLMediaElement*>( mpc );
	hme->play();

	//~ pthread_cond_signal( &This->m_ProcessSwitchCond );
    //~ pthread_mutex_unlock( &This->m_ProcessSwitchMutex );
}

void MediaPlayerPrivateMobiclip::ProcessSwitchAbortStatic(void* context)
{
	MediaPlayerPrivateMobiclip* This = static_cast<MediaPlayerPrivateMobiclip*>(context);
    pthread_mutex_lock(&This->m_ProcessSwitchMutex);

    // Save the playback position to restore it if we come back in foreground
    // see OnStateChange
    This->m_fgCurrentTimeSecs = This->currentTime();
    This->m_iSeekPerformed = false;

    //
    // Use available method from cache management to freeze Webkit's HTLMMediaElement
    //
    HTMLMediaElement* mediaElement = static_cast<HTMLMediaElement*>( This->m_player->mediaPlayerClient() );
    Document* document = mediaElement->document();
    document->suspendActiveDOMObjects(ActiveDOMObject::DocumentWillBecomeInactive);

    //
    // Free all resources used by Mobiclip's mediaplayer (end all threads)
    //
    This->m_v->Abort();

    This->m_isInBackground = true;

	pthread_cond_signal( &This->m_ProcessSwitchCond );
    pthread_mutex_unlock(&This->m_ProcessSwitchMutex);
}

void MediaPlayerPrivateMobiclip::ProcessSwitchResumeState(void* context)
{
	MediaPlayerPrivateMobiclip* This = static_cast<MediaPlayerPrivateMobiclip*>(context);

    pthread_mutex_lock(&This->m_ProcessSwitchMutex);

    CString curl = This->m_mediaURL.utf8();
    const char* surl = curl.data();

    //
    // We have to restart Webkit's HTLMMediaElement
    // Before that Mobiclip's mediaplayer must be restored
    // See ProcessSwitchAbortStatic
    // See OnStateChange
    //
    This->m_v->SetSource( surl );
    This->m_v->Prepare(); // This call will load the video and put the mediaplayer in pause state

	pthread_cond_signal( &This->m_ProcessSwitchCond );
    pthread_mutex_unlock(&This->m_ProcessSwitchMutex);
}


void MediaPlayerPrivateMobiclip::ProcessSwitchLoop()
{
	// Register us for process switch notification
	//bool bBackground = shim_IsBackground();
	shim_RegisterSyncThread(NULL, NULL, NULL);

	pthread_mutex_lock(&m_ProcessSwitchMutex);
	bool paused_for_bg = false;
	for(bool done = false; !done;)
	{
		struct timeval now;
		gettimeofday(&now, NULL);
		struct timespec until;
		until.tv_sec  = now.tv_sec;
		until.tv_nsec = now.tv_usec * 1000;
		until.tv_nsec += 15 * 1e6;

		int e = pthread_cond_timedwait( &m_ProcessSwitchCond, &m_ProcessSwitchMutex, &until );
		if( e!=ETIMEDOUT )
			break;

		int switching = shim_PollForSwitch();
		switch(switching)
		{
			case SHIM_SWITCH_TO_BG:
				paused_for_bg = !paused();
				callOnMainThread( ProcessSwitchPauseStatic, this );
				//~ pthread_cond_wait( &m_ProcessSwitchCond, &m_ProcessSwitchMutex );
				//~ {
					//~ MediaPlayerClient* const mpc = m_player->mediaPlayerClient();
					//~ WebCore::HTMLMediaElement* const hme = static_cast<WebCore::HTMLMediaElement*>( mpc );
					//~ hme->pause();
				//~ }
				//~ WebKitPause();
				shim_AcknowledgeSwitch();
				break;
			case SHIM_SWITCH_SHUTDOWN:
				cancelLoad();
				shim_AcknowledgeSwitch();
				pthread_detach(pthread_self());
				done = true;
				break;
			case SHIM_SWITCH_TO_FG:
				if(paused_for_bg)
				{
					paused_for_bg = false;
					callOnMainThread( ProcessSwitchPlayStatic, this );
				}
				//~ pthread_cond_wait( &m_ProcessSwitchCond, &m_ProcessSwitchMutex );
				//~ {
					//~ MediaPlayerClient* const mpc = m_player->mediaPlayerClient();
					//~ WebCore::HTMLMediaElement* const hme = static_cast<WebCore::HTMLMediaElement*>( mpc );
					//~ hme->play();
				//~ }
				//~ WebKitResume();
				shim_AcknowledgeSwitch();
				break;
			default:
				break;
		}
	}
	pthread_mutex_unlock(&m_ProcessSwitchMutex);

	shim_UnregisterSyncThread(NULL);
}



}; // namespace WebCore

#endif
