/*
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions, and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY NINTENDO AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL NINTENDO OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*

	Video Playback for WiiU
	(high level interface with Webkit)

*/
#include "config.h"

#if ENABLE(VIDEO) && (PLATFORM(WIIU))

// --------------------------------------------------------------------

#include "MediaPlayerPrivateWiiU.h"
#include "TimeRanges.h"
#include "DomainAccessControlWiiU.h"
#include "GraphicsContext.h"
#include "GraphicsContextPlatformPrivate.h"
#include "Frame.h"
#include "FrameLoaderClientWiiU.h"
#include <PlatformError.h>

#if ENABLE_XSOCK
#include "xsockInterface.h"
#ifndef SYSTEMSLEEP
#define SYSTEMSLEEP(x) OSSleepMilliseconds(x);
#endif
#endif

using namespace std;

String	WebCore::MediaPlayerPrivateWiiU::s_lastQueryType;
String	WebCore::MediaPlayerPrivateWiiU::s_lastQueryCodecs;

#if ENABLE_XSOCK
class MEDIAtransportFileHook
{
public:
	virtual long long GetFileSize(int fd) = 0;
	virtual int Open(const char *url) = 0;
	virtual int Read(int fd, long long offset, void *buffer, int count) = 0;
	virtual void Close(int fd) = 0;
};
extern MEDIAtransportFileHook *MEDIAtransportFile_Hook;
#endif

namespace WebCore {
	
#if ENABLE_XSOCK
class MEDIAtransportFileHook_Xsock : public MEDIAtransportFileHook
{
	enum {
		PAGE_SIZE = 32768,
		MAX_PAGE_COUNT = 8
	};

	typedef std::vector<int> IdList;
	struct Page {
		int cmdId;
		int offset;
	};
	struct File {
		Page pages[MAX_PAGE_COUNT];
		int curPage;
		xsockInterface::NetworkFileHandle fd;
		xsockClient *socket;
		long long fileSize;
		long long curOffset;
		String uri;
		IdList discards;
	};
	typedef std::map<int, File> FileMap;

	int m_refCount;
	int m_nextFd;
	FileMap m_files;
	pthread_mutex_t m_filesMutex;
	char m_buffer[PAGE_SIZE];

	int getNextFd() {
		pthread_mutex_lock(&m_filesMutex);
		int ret = m_nextFd;
		if (++m_nextFd <= 0) m_nextFd = 1;
		if (m_files.count(ret)) {
			pthread_mutex_unlock(&m_filesMutex);
			return getNextFd();
		}
		pthread_mutex_unlock(&m_filesMutex);
		return ret;
	}

	int getNextPage(int curPage) {
		return (curPage + 1) % MAX_PAGE_COUNT;
	}

	long long getFileSizeLocked(File &file) {
		if (file.fileSize >= 0)
			return file.fileSize;

		xsockInterface::xsockCallbacks &callbacks = file.socket ? xsockInterface::networkCallbacks : xsockInterface::localCallbacks;
		long long result;
		if (callbacks.getFileSizeFunc) {
			if (callbacks.getFileSizeFunc(file.socket, file.uri.utf8().data(), &file.fileSize))
				return file.fileSize;
		}
		return -1;
	}

	int requestNextPage(File &file, long long offset) {
		if (!file.socket) return 0;

		long long remainingSize = getFileSizeLocked(file);
		remainingSize -= offset;
		if (remainingSize < 0) return -1;

		if (offset != file.curOffset) {
			discardAllPages(file);
			if (xsockInterface::networkCallbacks.seekFileFunc) {
				long long resultSeek;
				if ((resultSeek = xsockInterface::networkCallbacks.seekFileFunc(file.socket, file.fd, offset, 0)) != offset) {
					return -1;
				}
			}
			else return -1;
		}
		int firstAvailablePage = file.curPage;
		while (file.pages[firstAvailablePage].cmdId) {
			firstAvailablePage = getNextPage(firstAvailablePage);
			if (firstAvailablePage == file.curPage) return 0;
		}
		int numAvailablePages = (file.curPage - firstAvailablePage + MAX_PAGE_COUNT) % MAX_PAGE_COUNT;
		if (numAvailablePages == 0) numAvailablePages = MAX_PAGE_COUNT;
		int sizeToRead = numAvailablePages * PAGE_SIZE;
		if (remainingSize < sizeToRead) sizeToRead = remainingSize;
		int numPagesUsed = sizeToRead / PAGE_SIZE + (sizeToRead % PAGE_SIZE ? 1 : 0);
		if (numPagesUsed == 0) return 0;
		if (xsockInterface::networkCallbacks.readFromFileNetwork_asyncSend) {
			int cmd = xsockInterface::networkCallbacks.readFromFileNetwork_asyncSend(file.socket, file.fd, sizeToRead, PAGE_SIZE);
			if (cmd < 0) return -1;
			for (int i = 0; i < numPagesUsed; ++i) {
				file.pages[firstAvailablePage].cmdId = cmd + i;
				file.pages[firstAvailablePage].offset = offset + i * PAGE_SIZE;
				firstAvailablePage = getNextPage(firstAvailablePage);
			}
			file.curOffset = offset + sizeToRead;
			return numPagesUsed;
		}
		else return -1;
	}

	int discardPage(File &file, int page) {
		if (file.pages[page].cmdId == 0) return 0;
		file.discards.push_back(file.pages[page].cmdId);
		file.pages[page].cmdId = 0;
		file.pages[page].offset = 0;
		return 1;
	}

	void discardAllPages(File &file) {
		while (discardPage(file, file.curPage)) file.curPage = getNextPage(file.curPage);
	}

	int clearDiscardedPages(File &file) {
		if (!file.socket) return 0;
		size_t numDiscarded = 0;
		for (numDiscarded = 0; numDiscarded < file.discards.size(); ++numDiscarded) {
			int result = xsockInterface::networkCallbacks.readFromFileNetwork_asyncRecv(file.socket, file.discards[numDiscarded], NULL);
			if (result < 0) {
				if (result == xsockInterface::operationWouldBlock) break;
				file.discards.clear();
				return 0;
			}
		}
		if (numDiscarded) file.discards.erase(file.discards.begin(), file.discards.begin() + numDiscarded);

		return (int)file.discards.size();
	}

public:

	int AddRef() {
		pthread_mutex_lock(&m_filesMutex);
		int ret = ++m_refCount;
		pthread_mutex_unlock(&m_filesMutex);
		return ret;
	}

	int Release() {
		pthread_mutex_lock(&m_filesMutex);
		int ret = --m_refCount;
		pthread_mutex_unlock(&m_filesMutex);
		return ret;
	}

public:

	MEDIAtransportFileHook_Xsock() : m_refCount(0), m_nextFd(1) {
		pthread_mutex_init(&m_filesMutex, NULL);
		AddRef();
	}

	~MEDIAtransportFileHook_Xsock() {
		#if 0 // I bet this is unsafe if destruction happens after xsock is already cleaned up
		while (!m_files.empty()) {
			int fd = m_files.begin()->first;
			Close(fd);
		}
		#endif
		pthread_mutex_destroy(&m_filesMutex);
	}

	virtual long long GetFileSize(int fd) {
		pthread_mutex_lock(&m_filesMutex);
		if (!m_files.count(fd)) {
			pthread_mutex_unlock(&m_filesMutex);
			return -1;
		}

		File &file = m_files[fd];
		long long ret = getFileSizeLocked(file);
		pthread_mutex_unlock(&m_filesMutex);
		return ret;
	}

	virtual int Open(const char *uri) {
		xsockClient *socket = NULL;
		if (xsockInterface::localCallbacks.isNetworkHostFunc)
			xsockInterface::localCallbacks.isNetworkHostFunc("", &socket);
		xsockInterface::xsockCallbacks &callbacks = socket ? xsockInterface::networkCallbacks : xsockInterface::localCallbacks;
		int fd = getNextFd();
		pthread_mutex_lock(&m_filesMutex);

		File &file = m_files[fd];
		memset(&file.pages, 0, sizeof(Page) * MAX_PAGE_COUNT);
		file.curPage = 0;
		file.fd = -1;
		file.socket = socket;
		file.fileSize = -1;
		file.curOffset = 0;
		file.uri = uri;

		if (callbacks.openFileFunc)
			file.fd = callbacks.openFileFunc(socket, uri, "r");
		if (file.fd < 0) {
			m_files.erase(fd);
			fd = -1;
		}
		else {
			requestNextPage(file, 0);
		}

		pthread_mutex_unlock(&m_filesMutex);
		return fd;
	}

	virtual int Read(int fd, long long offset, void *buffer, int count) {
		pthread_mutex_lock(&m_filesMutex);
		if (!m_files.count(fd)) {
			pthread_mutex_unlock(&m_filesMutex);
			return 0;
		}
		File &file = m_files[fd];
		int sizeRead = 0;
		if (file.socket) {
			clearDiscardedPages(file);
			if (file.pages[file.curPage].cmdId && file.pages[file.curPage].offset == offset) {
				int read = xsockInterface::operationWouldBlock;
				while (read == xsockInterface::operationWouldBlock) {
					read = xsockInterface::networkCallbacks.readFromFileNetwork_asyncRecv(file.socket, file.pages[file.curPage].cmdId, m_buffer);
					if (read == xsockInterface::operationWouldBlock) {
						SYSTEMSLEEP(1);
					}
				}
				file.pages[file.curPage].cmdId = 0;
				file.pages[file.curPage].offset = 0;
				file.curPage = getNextPage(file.curPage);
				if (read > 0) {
					if (count < read) read = count;
					memcpy(buffer, (void*)m_buffer, read);
					sizeRead = read;
					requestNextPage(file, file.curOffset);
				}
			}
			else {
				discardAllPages(file);
				if (requestNextPage(file, offset) > 0) {
					pthread_mutex_unlock(&m_filesMutex);
					return Read(fd, offset, buffer, count);
				}
			}
		}
		else {
			if (file.curOffset != offset) {
				if (xsockInterface::localCallbacks.seekFileFunc) {
					file.curOffset = xsockInterface::localCallbacks.seekFileFunc(file.socket, file.fd, offset, 0);
				}
				if (file.curOffset != offset) {
					pthread_mutex_unlock(&m_filesMutex);
					return 0;
				}
			}

			if (xsockInterface::localCallbacks.readFromFileFunc) {
				while (count > 0) {
					int result = xsockInterface::localCallbacks.readFromFileFunc(file.socket, file.fd, ((char*)buffer) + sizeRead, count);
					if (result < 0) break;
					count -= result;
					sizeRead += result;
				}
				file.curOffset += sizeRead;
			}
		}
		pthread_mutex_unlock(&m_filesMutex);
		return sizeRead;
	}

	virtual void Close(int fd) {
		pthread_mutex_lock(&m_filesMutex);
		if (!m_files.count(fd)) {
			pthread_mutex_unlock(&m_filesMutex);
			return;
		}

		File &file = m_files[fd];
		xsockInterface::xsockCallbacks &callbacks = file.socket ? xsockInterface::networkCallbacks : xsockInterface::localCallbacks;
		if (callbacks.closeFileFunc)
			callbacks.closeFileFunc(file.socket, &file.fd);
		discardAllPages(file);
		while (clearDiscardedPages(file));
		callbacks.closeNetworkHostFunc(file.socket);
		m_files.erase(fd);

		pthread_mutex_unlock(&m_filesMutex);
	}

};
static MEDIAtransportFileHook_Xsock MEDIAtransportFileHook_instance;
#endif

// --------------------------------------------------------------------

// --------------------------------------------------------------------
/**
	Support function to not have to rebuild the MIME info over and over again
*/
static HashSet<String> mimeTypeCache()
	{
	DEFINE_STATIC_LOCAL(HashSet<String>, cache, ());
	static bool _bTypeListInitialized = false;

	if(!_bTypeListInitialized)
		{
		cache.add(String("audio/ogg"));						// Ogg Vorbis
#ifdef ENABLE_MPEG_DECODER
		cache.add(String("audio/mpeg"));					// Mpeg audio, layers 1/2/3
#endif		
		cache.add(String("audio/wav"));						// Windows WAV
		cache.add(String("audio/x-wav"));					// Windows WAV
		cache.add(String("audio/wave"));					// Windows WAV
		cache.add(String("application/vnd.apple.mpegurl"));	// HLS
		cache.add(String("vnd.apple.mpegurl"));
		cache.add(String("application/x-mpegurl"));			// HLS, too
		cache.add(String("audio/aac"));						// AAC
		cache.add(String("video/mp4"));						// H.264 & possibly AAC
		cache.add(String("video/mp2t"));					// MP2TS

		_bTypeListInitialized = true;
		}
	return cache;
	}

// --------------------------------------------------------------------
/**
	Register the supported MIME types of this media engine
*/
void MediaPlayerPrivateWiiU::getSupportedTypes(HashSet<String>& types)
{
	const HashSet<String>& mimeCache = mimeTypeCache();

	HashSet<String>::const_iterator it = mimeCache.begin();
    HashSet<String>::const_iterator end = mimeCache.end();
    for (; it != end; ++it )
            types.add( *it );
}

// -------------------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------------------

// --------------------------------------------------------------------
/**
	Create an instance of this media engine
*/
PassOwnPtr<MediaPlayerPrivateInterface> MediaPlayerPrivateWiiU::create(MediaPlayer* player)
{
	#if ENABLE_XSOCK
	if (MEDIAtransportFile_Hook == 0) MEDIAtransportFile_Hook = &MEDIAtransportFileHook_instance;
	#endif

    return adoptPtr(new MediaPlayerPrivateWiiU(player));
}

// --------------------------------------------------------------------
/**
	Check if this media engine is available
*/
bool MediaPlayerPrivateWiiU::isAvailable()
{
	// In this context we assume always available player
	return true;
}

// --------------------------------------------------------------------
/**
	Check if we support a given type
*/
MediaPlayer::SupportsType MediaPlayerPrivateWiiU::supportsType(const String& type, const String& codecs)
{
	s_lastQueryType = type;
	s_lastQueryCodecs = codecs;

	if (type.isNull() && type.isEmpty())
		return MediaPlayer::IsNotSupported;

    // spec says we should not return "probably" if the codecs string is empty
    if (mimeTypeCache().contains(type))
        return codecs.isEmpty() ? MediaPlayer::MayBeSupported : MediaPlayer::IsSupported;

	std::string typeCheck = type.utf8().data();
	if(typeCheck.find("audio") != std::string::npos)
	{
		LibPlatform::Error::GetInstance()->Report(LibPlatform::Error::ERROR_CODE_WEBKIT_UNSUPPORTED_AUDIO_FORMAT, L"", 0);
	}
	else
	{
		LibPlatform::Error::GetInstance()->Report(LibPlatform::Error::ERROR_CODE_WEBKIT_UNSUPPORTED_VIDEO_FORMAT, L"", 0);
	}

	return MediaPlayer::IsNotSupported;
}

// --------------------------------------------------------------------
/**
	Register the media engine with Webkit
*/
void MediaPlayerPrivateWiiU::registerMediaEngine(MediaEngineRegistrar registrar)
{
	if (isAvailable())
        registrar(create, getSupportedTypes, supportsType, 0, 0, 0);
}

// --------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateWiiU::doLoad(const String& url)
{
	m_url = encodeWithURLEscapeSequences(url).latin1().data();

	m_type = s_lastQueryType.utf8().data();
	m_codecs = s_lastQueryCodecs.utf8().data();

OS::Report("Load %s MIME=%s CODECS=%s\n", m_url.c_str(), m_type.c_str(), m_codecs.c_str());

	if (m_preLoadMode == MediaPlayer::Auto)
		openStream();
}

// --------------------------------------------------------------------------
/**
*/
bool MediaPlayerPrivateWiiU::isURLallowable(const std::string & url)
	{
			int error = WebCore::DomainAccessControl::DOMAIN_ACCESS_OK;
	if (!WebCore::DomainAccessControl::IsDomainAccessible(url.c_str(), &error))
			{
				switch (error)
			{
				case WebCore::DomainAccessControl::DOMAIN_ACCESS_NOT_WHITELISTED:
					OS::Report("The url '%s' is not whitelisted. Add its domain to config.xml to whitelist.\n", url.c_str());
					LibPlatform::Error::GetInstance()->Report(LibPlatform::Error::ERROR_CODE_WEBKIT_ADDRESS_BLOCKED_BY_WHITELIST, L"", 0);
					break;

				case WebCore::DomainAccessControl::DOMAIN_ACCESS_NOT_ACQUIRED_AUTH_TOKEN:
					OS::Report("Before accessing a network resource, an independent service token must first be acquired.\n" \
							"Please call nwf.act.NintendoAccountManager.acquireNewIndependentServiceToken().\n");
					break;

				default:
					OS::Report("This domain is unreachable. Error %d\n", error);
					break;
				}
				return false;
			}
	return true;
}

// --------------------------------------------------------------------------
/**
	Return index of display the player is on

	@return	Index of display (TV=0, GAMEPAD=1,2...); returns -1 if unknown
*/

int MediaPlayerPrivateWiiU::getDisplayIndex() const
{
	FrameLoader *pFrameLoader = m_player->mediaPlayerClient()->mediaPlayerOwningDocument()->frame()->loader();
	if (!pFrameLoader)
		return -1;

	WebView *pView = static_cast<FrameLoaderClientWiiU*>(pFrameLoader->client())->webView();

	if (pView->GetDisplayType() == WebView::GAMEPAD_DISPLAY_TYPE)
		return 1 + pView->GetGamePadIndex();
	return 0;
}

// --------------------------------------------------------------------------
/**
*/
Timer<MediaTimer>* MediaPlayerPrivateWiiU::CreateTimer(MediaTimer *pTimer)
{
	return new Timer<MediaTimer>(pTimer, &MediaTimer::_timerCallback);
}

// --------------------------------------------------------------------------
/**
*/
void MediaPlayerPrivateWiiU::DestroyTimer(Timer<MediaTimer>* pTimer)
{
	delete pTimer;
}

// --------------------------------------------------------------------------
/**
*/
void  MediaPlayerPrivateWiiU::StartTimer(Timer<MediaTimer>* pTimer, double period)
{
	pTimer->startRepeating(period);
}

// --------------------------------------------------------------------------
/**
*/
void  MediaPlayerPrivateWiiU::StopTimer(Timer<MediaTimer>* pTimer)
{
	pTimer->stop();
}

// --------------------------------------------------------------------------
/**
 */
void MediaTimer::_timerCallback(WebCore::Timer<MediaTimer>*)
{
	timerCallback();
}

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------
// - Interface for wksupport calls ------------------------------------------
// --------------------------------------------------------------------------
// --------------------------------------------------------------------------

// --------------------------------------------------------------------------
/**
*/
Timer<MediaTimer>* GraphicsContextPlatformPrivate::WKgfxHooks::MediaPlayerPrivateWiiU_CreateTimer(MediaTimer *pTimer)
{
	return MediaPlayerPrivateWiiU::CreateTimer(pTimer);
}

// --------------------------------------------------------------------------
/**
*/
void GraphicsContextPlatformPrivate::WKgfxHooks::MediaPlayerPrivateWiiU_DestroyTimer(Timer<MediaTimer>* pTimer)                                                                
{
	MediaPlayerPrivateWiiU::DestroyTimer(pTimer);
}

// --------------------------------------------------------------------------
/**
*/
void GraphicsContextPlatformPrivate::WKgfxHooks::MediaPlayerPrivateWiiU_StartTimer(Timer<MediaTimer>* pTimer, double period)                                                      
{
	MediaPlayerPrivateWiiU::StartTimer(pTimer, period);
}

// --------------------------------------------------------------------------
/**
*/
void GraphicsContextPlatformPrivate::WKgfxHooks::MediaPlayerPrivateWiiU_StopTimer(Timer<MediaTimer>* pTimer)                                                                   
{
	MediaPlayerPrivateWiiU::StopTimer(pTimer);
}

// --------------------------------------------------------------------------
/**
*/
int GraphicsContextPlatformPrivate::WKgfxHooks::MediaPlayerPrivateWiiU_getDisplayIndex(const MediaPlayerPrivateWiiU *pThis)                                                    
{
	return pThis->getDisplayIndex();
}

// --------------------------------------------------------------------------
/**
*/
bool GraphicsContextPlatformPrivate::WKgfxHooks::MediaPlayerPrivateWiiU_isURLallowable(const std::string & url)                                                                         
{
	return MediaPlayerPrivateWiiU::isURLallowable(url);
	}

// --------------------------------------------------------------------------
/**
*/
void GraphicsContextPlatformPrivate::WKgfxHooks::MediaPlayerPrivateWiiU_doLoad(MediaPlayerPrivateWiiU *pThis, const WTF::String & url)                                           
{
	pThis->doLoad(url);
}

// --------------------------------------------------------------------------
/**
*/
void GraphicsContextPlatformPrivate::WKgfxHooks::MediaPlayer_networkStateChanged(MediaPlayer *pThis)                                                                           
{
	pThis->networkStateChanged();
}

// --------------------------------------------------------------------------
/**
*/
void GraphicsContextPlatformPrivate::WKgfxHooks::MediaPlayer_readyStateChanged(MediaPlayer *pThis)                                                                             
{
	pThis->readyStateChanged();
}

// --------------------------------------------------------------------------
/**
*/
void GraphicsContextPlatformPrivate::WKgfxHooks::MediaPlayer_timeChanged(MediaPlayer *pThis)                                                                                   
	{
	pThis->timeChanged();
}

// --------------------------------------------------------------------------
/**
*/
void GraphicsContextPlatformPrivate::WKgfxHooks::MediaPlayer_repaint(MediaPlayer *pThis)                                                                                       
{
	pThis->repaint();
}

// --------------------------------------------------------------------------
/**
*/
void GraphicsContextPlatformPrivate::WKgfxHooks::GraphicsContext_fillRect(GraphicsContext *pThis, const FloatRect& r, const Color& c, ColorSpace cs)
{
	pThis->fillRect(r, c, cs);
}

// --------------------------------------------------------------------------
/**
*/
bool GraphicsContextPlatformPrivate::WKgfxHooks::GraphicsContext_paintingDisabled(const GraphicsContext *pThis)                                                                
{
	return pThis->paintingDisabled();
}

// --------------------------------------------------------------------------
/**
*/
const PlatformMedia & GraphicsContextPlatformPrivate::WKgfxHooks::getNoPlatformMedia()
{
	return getNoPlatformMedia();
}

}; // namespace WebCore


#endif

