/*
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions, and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY NINTENDO AND ITS CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL NINTENDO OR CONTRIBUTORS BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "JSEventSystem.h"
#include <wtf/Platform.h> // PLATFORM(...)
#include <PlatformOS.h>
using namespace LibPlatform;

JSEvent::JSEvent(WiiUEventType_t _event_id, std::string event_type, void *data, std::string event_string) :
	event_type(event_type),
	event_string(event_string),
	data(data),
	event_id(_event_id) {
}

JSEvent::~JSEvent(void) {
}

std::string JSEvent::ToString(void) {
	return this->event_type;
}

JSEventListener::JSEventListener(std::string listener_name) :
	listener_name(listener_name) {
}

JSEventListener::~JSEventListener(void) {
}

std::string JSEventListener::ToString(void) {
	return this->listener_name;
}

//void JSEventListener::AddListenerCallback(JSEventListenerData* callback)
//{
//	listener_callbacks.push_back(callback);
//}
//
//void JSEventListener::ClearListenerCallbacks(void)
//{
//	for(int i = 0; i < listener_callbacks.size(); ++i) {
//		delete listener_callbacks[i];
//	}
//
//	listener_callbacks.clear();
//}

JSEventDispatcher::JSEventDispatcher(std::string dispatcher_name) : 
	dispatcher_name(dispatcher_name) {
}

JSEventDispatcher::~JSEventDispatcher(void) {
	std::map<std::string, std::vector<JSEventListener *> *>::iterator i;
	for(i = this->event_listener_map.begin(); i != this->event_listener_map.end(); ++i) {
		if(i->second) delete i->second;
	}
}

bool JSEventDispatcher::AddEventListener(JSEventListener *js_listener, std::string event_type) {
	std::map<std::string, std::vector<JSEventListener *> *>::const_iterator mit = this->event_listener_map.find(event_type);
	std::vector<JSEventListener *> *listeners = NULL;
	
	// If there currently are no listeners for the given event type, then create a new list.
	// If there are make sure this dispatcher is not already listening to this event.
	if(mit == this->event_listener_map.end()) {
		listeners = new std::vector<JSEventListener *>();
		this->event_listener_map[event_type] = listeners;
	} else {
		listeners = mit->second;
		for(std::vector<JSEventListener *>::iterator i = listeners->begin(); i != listeners->end(); ++i) {
			if((*i) == js_listener) {
				return false;
			}
		}
	}

	listeners->push_back(js_listener);
	return true;
}

void JSEventDispatcher::RemoveEventListener(JSEventListener *js_listener, std::string event_type) {
	std::map<std::string, std::vector<JSEventListener *> *>::const_iterator mit = this->event_listener_map.find(event_type);
	
	// If there are listeners on the given type then try to find the desired one to remove.
	if(mit != this->event_listener_map.end()) {
		std::vector<JSEventListener *>* listeners = (mit->second);
		// Loop until we find the event listener or the end of the list.
		unsigned int index = 0;
		while((index < listeners->size()) && ((*listeners)[index] != js_listener)) {
			OS::Report("%u %p == %p\n", index, (*listeners)[index], js_listener);
			++index;
		}

		// If we found the listener replace with the last listener of the list and pop the end.
		if(index < listeners->size() && (*listeners)[index] == js_listener) {
			(*listeners)[index] = (*listeners)[listeners->size() - 1];
			(*listeners).pop_back();
		}
	}
}

void JSEventDispatcher::RemoveAllListeners(JSEventListener *js_listener) {

	std::map<std::string, std::vector<JSEventListener *> *>::iterator it = event_listener_map.begin();

	//remove js_listener from all possible event types
	for(; it != event_listener_map.end(); ++it)
	{
		std::vector<JSEventListener *> *listeners = this->event_listener_map[it->first];

		if(listeners)
		{
			for(int i = 0; i < listeners->size(); ++i)
			{
				if((*listeners)[i] == js_listener)
				{
					RemoveEventListener(js_listener, it->first);
				}
			}
		}
	}
}


void JSEventDispatcher::DispatchEvent(JSEvent js_event) {
	DispatchEvent(js_event, js_event.ToString());
}

void JSEventDispatcher::DispatchEvent(JSEvent js_event, std::string event_string) {
	// Retrieve the list of listeners for the given event's type.
	std::map<std::string, std::vector<JSEventListener *> *>::const_iterator mit = this->event_listener_map.find(event_string);

	// Send given event to all listeners of its type.
	if(mit != this->event_listener_map.end()) {
		// Creating a copy of the vector here is horribly inefficient. However, it is necessary.
		// It is possible for the JS callback that is called by the below call to OnEvent() to
		// call addEventListener, which adds a listener to event_listener_map (all in the same
		// thread, so mutexes won't work). If event_listener_map gets event listeners added to
		// it while it's being iterated over right now, crash.
		std::vector<JSEventListener *> listeners = *(mit->second);
		for(std::vector<JSEventListener *>::iterator i = listeners.begin(); i != listeners.end(); ++i) {
			(*i)->OnEvent(js_event);
		}
	}
}

unsigned int JSEventDispatcher::HasEventListener(std::string event_type) {
	// If we have any listeners for the given event return how many.
	std::map<std::string, std::vector<JSEventListener *> *>::const_iterator mit = this->event_listener_map.find(event_type);
	if(mit != this->event_listener_map.end()) return mit->second->size();
	else return 0;
}

bool JSEventDispatcher::HasEventListener(JSEventListener *js_listener, std::string event_type) {
	std::map<std::string, std::vector<JSEventListener *> *>::const_iterator mit = this->event_listener_map.find(event_type);
	
	// No listeners for this event type
	if(mit == this->event_listener_map.end()) return false;

	// Loop until we find the event listener or the end of the list.
	unsigned int index = 0;
	while((index < mit->second->size()) && ((*mit->second)[index] != js_listener)) ++index;

	// If we found the listener before the end of the list return true, otherwise false
	return (index < mit->second->size() && (*mit->second)[index] == js_listener);
}

std::string JSEventDispatcher::ToString(void) {
	// First print dispatcher's name.
	std::string dispatcher_string(this->dispatcher_name);
	dispatcher_string += "\n";
	
	// Loop through all map pairs and print out the event and all the listeners for that event.
	// For example: "EventType: Listener0 Listener1 ... ListenerN"
	std::map<std::string, std::vector<JSEventListener *> *>::iterator i;
	for(i = this->event_listener_map.begin(); i != this->event_listener_map.end(); ++i) {
		// Print event type.
		dispatcher_string += i->first + ": ";

		// Print all listeners.
		if(i->second) {
			for(std::vector<JSEventListener *>::iterator j = i->second->begin(); j != i->second->end(); ++j) {
				dispatcher_string += (*j)->ToString() + " ";
			}
		}

		dispatcher_string += "\n";
	}

	return dispatcher_string;
}
